'''
Extension for enhancing sphinx documentation generation for cython module
'''

import re
import types
import sys
from os.path import dirname, join
import sphinx
from sphinx.ext.autodoc import MethodDocumenter

class CythonMethodDocumenter(MethodDocumenter):
    # XXX i don't understand the impact of having a priority more than the
    # attribute or instance method but the things is, if it's a cython module,
    # the attribute will be prefer over method.
    priority = 12

def is_cython_extension(what, obj):
    # try to check if the first line of the doc is a signature
    doc = obj.__doc__
    if not doc:
        return False
    doc = doc.split('\n')
    if not len(doc):
        return False
    doc = doc[0]

    # test for cython cpdef
    if what in ('attribute', 'method') and hasattr(obj, '__objclass__'):
        if not re.match('^([a-zA-Z_][a-zA-Z0-9_]*)\.([a-zA-Z_][a-zA-Z0-9_]*)\((.*)\)', doc):
            return False
        return True
    # test for cython class
    if what == 'class' and hasattr(obj, '__pyx_vtable__'):
        if not re.match('^([a-zA-Z_][a-zA-Z0-9_]*)\((.*)\)', doc):
            return False
        return True
    # test for python method in cython class
    if what in ('method', 'function') and obj.__class__ == types.BuiltinFunctionType:
        if not re.match('^([a-zA-Z_][a-zA-Z0-9_]*)\((.*)\)', doc):
            return False
        return True

def callback_docstring(app, what, name, obj, options, lines):
    if what == 'module':
        # remove empty lines
        while len(lines):
            line = lines[0].strip()
            if not line.startswith('.. _') and line != '':
                break
            lines.pop(0)

        # if we still have lines, remove the title
        if len(lines):
            lines.pop(0)

        # if the title is followed by a separator, remove it.
        if len(lines) and lines[0].startswith('=='):
            lines.pop(0)

    elif is_cython_extension(what, obj) and lines:
        if what == 'class':
            lines.pop(0)
        line = lines.pop(0)

        # trick to realign the first line to the second one.
        # FIXME: fail if we finishing with::
        line_with_text = [x for x in lines if len(x.strip())]
        if len(line_with_text) and line is not None and len(lines):
            l = len(line_with_text[0]) - len(line_with_text[0].lstrip())
        else:
            l = 0
        lines.insert(0, ' ' * l + line)

        # calculate the minimum space available
        min_space = 999
        for line in lines:
            if not line.strip():
                continue
            min_space = min(min_space, len(line) - len(line.lstrip()))

        # remove that kind of space now.
        if min_space > 0:
            spaces = ' ' * min_space
            for idx, line in enumerate(lines):
                if not line.strip():
                    continue
                if not line.startswith(spaces):
                    continue
                lines[idx] = line[min_space:]

def callback_signature(app, what, name, obj, options, signature,
                       return_annotation):
    # remove the first 'self' argument, because python autodoc don't
    # add it for python method class. So do the same for cython class.
    if is_cython_extension(what, obj):
        try:
            doc = obj.__doc__.split('\n').pop(0)
            doc = '(%s' % doc.split('(')[1]
            doc = doc.replace('(self, ', '(')
            doc = doc.replace('(self)', '( )')
            return (doc, None)
        except AttributeError:
            pass
        except IndexError:
            pass

def setup(app):
    import kivy
    sys.path += [join(dirname(kivy.__file__), 'extras')]
    from highlight import KivyLexer

    if sphinx.version_info[0] >= 3:
        app.add_lexer('kv', KivyLexer)
    else:
        app.add_lexer('kv', KivyLexer())
    app.add_autodocumenter(CythonMethodDocumenter)
    app.connect('autodoc-process-docstring', callback_docstring)
    app.connect('autodoc-process-signature', callback_signature)



'''
Canvas
======

The :class:`Canvas` is the root object used for drawing by a
:class:`~kivy.uix.widget.Widget`. Check the class documentation for more
information about the usage of Canvas.
'''

__all__ = ('Instruction', 'InstructionGroup',
           'ContextInstruction', 'VertexInstruction',
           'Canvas', 'CanvasBase',
           'RenderContext', 'Callback')

include "../include/config.pxi"
include "opcodes.pxi"

from kivy.graphics.cgl cimport *
from kivy.compat import PY2
from kivy.logger import Logger
from kivy.graphics.context cimport get_context, Context
from weakref import proxy
from threading import get_ident


cdef int _need_reset_gl = 1
cdef int _active_texture = -1
cdef list canvas_list = []

cdef void reset_gl_context():
    global _need_reset_gl, _active_texture
    _need_reset_gl = 0
    _active_texture = 0
    cgl.glEnable(GL_BLEND)
    cgl.glDisable(GL_DEPTH_TEST)
    cgl.glEnable(GL_STENCIL_TEST)
    cgl.glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
    cgl.glBlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_ONE, GL_ONE)
    cgl.glActiveTexture(GL_TEXTURE0)
    cgl.glPixelStorei(GL_UNPACK_ALIGNMENT, 1)


cdef class Instruction(ObjectWithUid):
    '''Represents the smallest instruction available. This class is for internal
    usage only, don't use it directly.
    '''
    def __cinit__(self):
        self.__proxy_ref = None
        self.flags = 0
        self.parent = None

    def __init__(self, **kwargs):
        self.group = kwargs.get('group', None)
        if kwargs.get('noadd'):
            self.flags |= GI_NO_REMOVE
            return

        if verify_gl_main_thread and initialized_tid \
                and get_ident() != initialized_tid:
            raise TypeError("Cannot create graphics instruction outside "
                            "the main Kivy thread")

        self.parent = getActiveCanvas()
        if self.parent:
            self.parent.add(self)

    cdef int apply(self) except -1:
        return 0

    IF DEBUG:
        cpdef flag_update(self, int do_parent=1, list _instrs=None):
            cdef list instrs = _instrs if _instrs else []
            if _instrs and self in _instrs:
                raise RuntimeError('Encountered instruction group render loop: %r in %r' % (self, _instrs,))
            if do_parent == 1 and self.parent is not None:
                instrs.append(self)
                self.parent.flag_update(do_parent=1, _instrs=instrs)
            self.flags |= GI_NEEDS_UPDATE
    ELSE:
        cpdef flag_update(self, int do_parent=1):
            if do_parent == 1 and self.parent is not None:
                self.parent.flag_update()
            self.flags |= GI_NEEDS_UPDATE

    cpdef flag_data_update(self):
        if verify_gl_main_thread and initialized_tid \
                and get_ident() != initialized_tid:
            raise TypeError("Cannot change graphics instruction outside "
                            "the main Kivy thread")
        self.flag_update()

    cdef void flag_update_done(self):
        self.flags &= ~GI_NEEDS_UPDATE

    cdef void radd(self, InstructionGroup ig):
        ig.children.append(self)
        self.set_parent(ig)

    cdef void rremove(self, InstructionGroup ig):
        if self.parent is None:
            return
        ig.children.remove(self)
        self.set_parent(None)

    cdef void rinsert(self, InstructionGroup ig, int index):
        ig.children.insert(index, self)
        self.set_parent(ig)

    cdef void set_parent(self, Instruction parent):
        self.parent = parent

    cdef void reload(self) except *:
        self.flags |= GI_NEEDS_UPDATE
        self.flags &= ~GI_NO_APPLY_ONCE
        self.flags &= ~GI_IGNORE

    @property
    def needs_redraw(self):
        if (self.flags & GI_NEEDS_UPDATE) > 0:
            return True
        return False

    @property
    def proxy_ref(self):
        '''Return a proxy reference to the Instruction i.e. without creating a
        reference of the widget. See `weakref.proxy
        <http://docs.python.org/2/library/weakref.html?highlight=proxy#weakref.proxy>`_
        for more information.

        .. versionadded:: 1.7.2
        '''
        if self.__proxy_ref is None:
            self.__proxy_ref = proxy(self)
        return self.__proxy_ref


cdef class InstructionGroup(Instruction):
    """
    Group of :class:`Instruction`. Allows for the adding and removing
    of graphics instructions. It can be used directly as follows::

        blue = InstructionGroup()
        blue.add(Color(0, 0, 1, 0.2))
        blue.add(Rectangle(pos=self.pos, size=(100, 100)))

        green = InstructionGroup()
        green.add(Color(0, 1, 0, 0.4))
        green.add(Rectangle(pos=(100, 100), size=(100, 100)))

        # Here, self should be a Widget or subclass
        [self.canvas.add(group) for group in [blue, green]]
    """
    def __init__(self, **kwargs):
        Instruction.__init__(self, **kwargs)
        self.children = list()
        self.compiled_children = None
        if 'nocompiler' in kwargs:
            self.compiler = None
        else:
            self.compiler = GraphicsCompiler()

    cdef int apply(self) except -1:
        cdef Instruction c
        cdef list children
        if self.compiler is not None:
            if self.flags & GI_NEEDS_UPDATE:
                self.build()
            if self.compiled_children is not None and not (self.flags & GI_NO_APPLY_ONCE):
                children = self.compiled_children.children
                for c in children:
                    if c.flags & GI_IGNORE:
                        continue
                    c.apply()
            self.flags &= ~GI_NO_APPLY_ONCE
        else:
            for c in self.children:
                c.apply()
        return 0

    cdef void build(self):
        self.compiled_children = self.compiler.compile(self)
        self.flag_update_done()

    cpdef add(self, Instruction c):
        '''Add a new :class:`Instruction` to our list.
        '''
        c.radd(self)
        self.flag_data_update()
        return

    cpdef insert(self, int index, Instruction c):
        '''Insert a new :class:`Instruction` into our list at index.
        '''
        c.rinsert(self, index)
        self.flag_data_update()

    cpdef remove(self, Instruction c):
        '''Remove an existing :class:`Instruction` from our list.
        '''
        c.rremove(self)
        self.flag_data_update()

    def indexof(self, Instruction c):
        cdef int i
        for i in xrange(len(self.children)):
            if self.children[i] is c:
                return i
        return -1

    def length(self):
        return len(self.children)

    cpdef clear(self):
        '''Remove all the :class:`Instructions <Instruction>`.
        '''
        cdef Instruction c
        for c in self.children[:]:
            if c.flags & GI_NO_REMOVE:
                continue
            self.remove(c)

    cpdef remove_group(self, str groupname):
        '''Remove all :class:`Instructions <Instruction>` with a specific group
        name.
        '''
        cdef Instruction c
        for c in self.children[:]:
            if c.flags & GI_NO_REMOVE:
                continue
            if c.group == groupname:
                self.remove(c)

    cpdef get_group(self, str groupname):
        '''Return an iterable for all the :class:`Instructions <Instruction>`
        with a specific group name.
        '''
        cdef Instruction c
        return [c for c in self.children if c.group == groupname]

    cdef void reload(self) except *:
        Instruction.reload(self)
        cdef Instruction c
        for c in self.children:
            c.reload()


cdef class ContextInstruction(Instruction):
    '''The ContextInstruction class is the base for the creation of instructions
    that don't have a direct visual representation, but instead modify the
    current Canvas' state, e.g. texture binding, setting color parameters,
    matrix manipulation and so on.
    '''
    def __init__(self, **kwargs):
        Instruction.__init__(self, **kwargs)
        self.flags |= GI_CONTEXT_MOD
        self.context_state = dict()
        self.context_push = list()
        self.context_pop = list()

    cdef RenderContext get_context(self):
        cdef RenderContext context = getActiveContext()
        return context

    cdef int apply(self) except -1:
        cdef RenderContext context = self.get_context()
        if self.context_push:
            context.push_states(self.context_push)
        if self.context_state:
            context.set_states(self.context_state)
        if self.context_pop:
            context.pop_states(self.context_pop)
        return 0

    cdef int set_state(self, str name, value) except -1:
        self.context_state[name] = value
        self.flag_update()

    cdef int push_state(self, str name) except -1:
        self.context_push.append(name)
        self.flag_update()

    cdef int pop_state(self, str name) except -1:
        self.context_pop.append(name)
        self.flag_update()

cdef class VertexInstruction(Instruction):
    '''The VertexInstruction class is the base for all graphics instructions
    that have a direct visual representation on the canvas, such as Rectangles,
    Triangles, Lines, Ellipse and so on.
    '''
    def __init__(self, **kwargs):
        # Set a BindTexture instruction to bind the texture used for
        # this instruction before the actual vertex instruction
        self.texture_binding = BindTexture(noadd=True, **kwargs)
        self.texture = self.texture_binding.texture #auto compute tex coords
        tex_coords = kwargs.get('tex_coords')
        if tex_coords:
            self.tex_coords = tex_coords

        Instruction.__init__(self, **kwargs)
        self.flags = GI_VERTEX_DATA & GI_NEEDS_UPDATE
        self.batch = VertexBatch()

    cdef void radd(self, InstructionGroup ig):
        cdef Instruction instr = self.texture_binding
        ig.children.append(self.texture_binding)
        ig.children.append(self)
        instr.set_parent(ig)
        self.set_parent(ig)

    cdef void rinsert(self, InstructionGroup ig, int index):
        cdef Instruction instr = self.texture_binding
        ig.children.insert(index, self.texture_binding)
        ig.children.insert(index, self)
        instr.set_parent(ig)
        self.set_parent(ig)

    cdef void rremove(self, InstructionGroup ig):
        cdef Instruction instr = self.texture_binding
        ig.children.remove(self.texture_binding)
        ig.children.remove(self)
        instr.set_parent(None)
        self.set_parent(None)

    @property
    def texture(self):
        '''Property that represents the texture used for drawing this
        Instruction. You can set a new texture like this::

            from kivy.core.image import Image

            texture = Image('logo.png').texture
            with self.canvas:
                Rectangle(texture=texture, pos=self.pos, size=self.size)

        Usually, you will use the :attr:`source` attribute instead of the
        texture.
        '''
        return self.texture_binding.texture

    @texture.setter
    def texture(self, _tex):
        cdef Texture tex = _tex
        self.texture_binding.texture = tex
        if tex:
            self.tex_coords = tex.tex_coords
        else:
            self.tex_coords = [0.0,0.0, 1.0,0.0, 1.0,1.0, 0.0,1.0]
        self.flag_data_update()

    @property
    def source(self):
        '''This property represents the filename to load the texture from.
        If you want to use an image as source, do it like this::

            with self.canvas:
                Rectangle(source='mylogo.png', pos=self.pos, size=self.size)

        Here's the equivalent in Kivy language:

        .. code-block:: kv

            <MyWidget>:
                canvas:
                    Rectangle:
                        source: 'mylogo.png'
                        pos: self.pos
                        size: self.size

        .. note::

            The filename will be searched for using the
            :func:`kivy.resources.resource_find` function.

        '''
        return self.texture_binding.source

    @source.setter
    def source(self, source):
        self.texture_binding.source = source
        self.texture = self.texture_binding._texture

    @property
    def tex_coords(self):
        '''This property represents the texture coordinates used for drawing the
        vertex instruction. The value must be a list of 8 values.

        A texture coordinate has a position (u, v), and a size (w, h). The size
        can be negative, and would represent the 'flipped' texture. By default,
        the tex_coords are::

            [u, v, u + w, v, u + w, v + h, u, v + h]

        You can pass your own texture coordinates if you want to achieve fancy
        effects.

        .. warning::

            The default values just mentioned can be negative. Depending
            on the image and label providers, the coordinates are flipped
            vertically because of the order in which the image is internally
            stored. Instead of flipping the image data, we are just flipping
            the texture coordinates to be faster.

        '''
        return (
            self._tex_coords[0],
            self._tex_coords[1],
            self._tex_coords[2],
            self._tex_coords[3],
            self._tex_coords[4],
            self._tex_coords[5],
            self._tex_coords[6],
            self._tex_coords[7])

    @tex_coords.setter
    def tex_coords(self, tc):
        cdef int index
        for index in xrange(8):
            self._tex_coords[index] = tc[index]
        self.flag_data_update()

    cdef void build(self):
        pass

    cdef int apply(self) except -1:
        if self.flags & GI_NEEDS_UPDATE:
            self.build()
            self.flag_update_done()
        self.batch.draw()
        return 0


cdef class Callback(Instruction):
    '''.. versionadded:: 1.0.4

    A Callback is an instruction that will be called when the drawing
    operation is performed. When adding instructions to a canvas, you can do
    this::

        with self.canvas:
            Color(1, 1, 1)
            Rectangle(pos=self.pos, size=self.size)
            Callback(self.my_callback)

    The definition of the callback must be::

        def my_callback(self, instr):
            print('I have been called!')

    .. warning::

        Note that if you perform many and/or costly calls to callbacks, you
        might potentially slow down the rendering performance significantly.

    The updating of your canvas does not occur until something new happens.
    From your callback, you can ask for an update::

        with self.canvas:
            self.cb = Callback(self.my_callback)
        # then later in the code
        self.cb.ask_update()

    If you use the Callback class to call rendering methods of another
    toolkit, you will have issues with the OpenGL context. The OpenGL state may
    have been manipulated by the other toolkit, and as soon as program flow
    returns to Kivy, it will just break. You can have glitches, crashes, black
    holes might occur, etc.
    To avoid that, you can activate the :attr:`reset_context` option. It will
    reset the OpenGL context state to make Kivy's rendering correct after the
    call to your callback.

    .. warning::

        The :attr:`reset_context` is not a full OpenGL reset. If you have issues
        regarding that, please contact us.

    '''
    def __init__(self, callback=None, **kwargs):
        Instruction.__init__(self, **kwargs)
        self.func = callback
        self._reset_context = int(kwargs.get('reset_context', False))

    def ask_update(self):
        '''Inform the parent canvas that we'd like it to update on the next
        frame. This is useful when you need to trigger a redraw due to some
        value having changed for example.

        .. versionadded:: 1.0.4
        '''
        self.flag_data_update()

    cdef int apply(self) except -1:
        cdef RenderContext rcx
        cdef Context ctx
        cdef Shader shader
        cdef int i

        cgl.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0)
        cgl.glBindBuffer(GL_ARRAY_BUFFER, 0)

        func = self.func
        if func is None or func(self):
            self.flag_update_done()

        if func is not None and self._reset_context:
            # FIXME do that in a proper way
            cgl.glDisable(GL_DEPTH_TEST)
            cgl.glDisable(GL_CULL_FACE)
            cgl.glDisable(GL_SCISSOR_TEST)
            cgl.glEnable(GL_BLEND)
            cgl.glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
            cgl.glBlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_ONE, GL_ONE)
            cgl.glUseProgram(0)

            # FIXME don't use 10. use max texture available from gl conf
            for i in xrange(10):
                cgl.glActiveTexture(GL_TEXTURE0 + i)
                cgl.glBindTexture(GL_TEXTURE_2D, 0)
                cgl.glDisableVertexAttribArray(i)
                cgl.glBindBuffer(GL_ARRAY_BUFFER, 0)
                cgl.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0)

            # reset all the vertexformat in all shaders
            ctx = get_context()
            for obj in ctx.lr_shader:
                shader = obj()
                if not shader:
                    continue
                shader.bind_vertex_format(None)

            # force binding again all our textures.
            rcx = getActiveContext()
            shader = rcx._shader
            rcx.enter()
            for index, texture in rcx.bind_texture.iteritems():
                rcx.set_texture(index, texture)

            reset_gl_context()
        return 0

    cdef int enter(self) except -1:
        self._shader.use()
        return 0

    @property
    def reset_context(self):
        '''Set this to True if you want to reset the OpenGL context for Kivy
        after the callback has been called.
        '''
        return self._reset_context

    @reset_context.setter
    def reset_context(self, value):
        cdef int ivalue = int(value)
        if self._reset_context == ivalue:
            return
        self._reset_context = ivalue
        self.flag_data_update()

    @property
    def callback(self):
        '''Property for getting/setting func.
        '''
        return self.func

    @callback.setter
    def callback(self, object func):
        if self.func == func:
            return
        self.func = func
        self.flag_data_update()


cdef class CanvasBase(InstructionGroup):
    '''CanvasBase provides the context manager methods for the
    :class:`Canvas`.'''
    def __enter__(self):
        pushActiveCanvas(self)

    def __exit__(self, *largs):
        popActiveCanvas()


cdef class Canvas(CanvasBase):
    '''The important Canvas class. Use this class to add graphics or context
    instructions that you want to be used for drawing.

    .. note::

        The Canvas supports Python's ``with`` statement and its enter & exit
        semantics.

    Usage of a canvas without the ``with`` statement::

        self.canvas.add(Color(1., 1., 0))
        self.canvas.add(Rectangle(size=(50, 50)))

    Usage of a canvas with Python's ``with`` statement::

        with self.canvas:
            Color(1., 1., 0)
            Rectangle(size=(50, 50))
    '''

    def __init__(self, **kwargs):
        get_context().register_canvas(self)
        CanvasBase.__init__(self, **kwargs)
        self._opacity = kwargs.get('opacity', 1.0)
        self._before = None
        self._after = None

    cdef void reload(self) except *:
        return
        '''
        # XXX ensure it's not needed anymore.
        cdef Canvas c
        if self._before is not None:
            c = self._before
            c.reload()
        CanvasBase.reload(self)
        if self._after is not None:
            c = self._after
            c.reload()
        '''

    cpdef clear(self):
        '''Clears every :class:`Instruction` in the canvas, leaving it clean.'''
        cdef Instruction c
        for c in self.children[:]:
            if c is self._before or c is self._after:
                continue
            if c.flags & GI_NO_REMOVE:
                continue
            self.remove(c)

    cpdef draw(self):
        '''Apply the instruction to our window.
        '''
        self.apply()

    cdef int apply(self) except -1:
        cdef float opacity = self._opacity
        cdef float rc_opacity
        cdef RenderContext rc
        if opacity != 1.0:
            rc = getActiveContext()
            rc_opacity = rc['opacity']
            rc.push_state('opacity')
            rc['opacity'] = rc_opacity * opacity
        InstructionGroup.apply(self)
        if opacity != 1.0:
            rc.pop_state('opacity')
        return 0

    cpdef add(self, Instruction c):
        # the after group must remain the last one.
        if self._after is None:
            c.radd(self)
        else:
            c.rinsert(self, -1)
        self.flag_data_update()

    cpdef remove(self, Instruction c):
        c.rremove(self)
        self.flag_data_update()

    def ask_update(self):
        '''Inform the canvas that we'd like it to update on the next frame.
        This is useful when you need to trigger a redraw due to some value
        having changed for example.
        '''
        self.flag_data_update()

    @property
    def before(self):
        '''Property for getting the 'before' group.
        '''
        if self._before is None:
            self._before = CanvasBase()
            self.insert(0, self._before)
        return self._before

    @property
    def after(self):
        '''Property for getting the 'after' group.
        '''
        cdef CanvasBase c
        if self._after is None:
            c = CanvasBase()
            self.add(c)
            self._after = c
        return self._after

    @property
    def has_before(self):
        '''Property to see if the :attr:`before` group has already been created.

        .. versionadded:: 1.7.0
        '''
        return self._before is not None

    @property
    def has_after(self):
        '''Property to see if the :attr:`after` group has already been created.

        .. versionadded:: 1.7.0
        '''
        return self._after is not None


    @property
    def opacity(self):
        '''Property to get/set the opacity value of the canvas.

        .. versionadded:: 1.4.1

        The opacity attribute controls the opacity of the canvas and its
        children.  Be careful, it's a cumulative attribute: the value is
        multiplied to the current global opacity and the result is applied to
        the current context color.

        For example: if your parent has an opacity of 0.5 and a child has an
        opacity of 0.2, the real opacity of the child will be 0.5 * 0.2 = 0.1.

        Then, the opacity is applied on the shader as::

            frag_color = color * vec4(1.0, 1.0, 1.0, opacity);

        '''
        return self._opacity

    @opacity.setter
    def opacity(self, value):
        self._opacity = value
        self.flag_data_update()

# Active Canvas and getActiveCanvas function is used
# by instructions, so they know which canvas to add
# themselves to
cdef CanvasBase ACTIVE_CANVAS = None

cdef CanvasBase getActiveCanvas():
    global ACTIVE_CANVAS
    return ACTIVE_CANVAS

# Canvas Stack, for internal use so canvas can be bound
# inside other canvas, and restored when other canvas is done
cdef list CANVAS_STACK = list()

cdef pushActiveCanvas(CanvasBase c):
    global ACTIVE_CANVAS, CANVAS_STACK
    CANVAS_STACK.append(ACTIVE_CANVAS)
    ACTIVE_CANVAS = c

cdef popActiveCanvas():
    global ACTIVE_CANVAS, CANVAS_STACK
    ACTIVE_CANVAS = CANVAS_STACK.pop()


#TODO: same as canvas, move back to context.pyx..fix circular import
#on actual import from python problem
include "common.pxi"
from .vertex cimport *
#from .texture cimport *

from os.path import join
from kivy import kivy_shader_dir
from kivy.cache import Cache
from kivy.core.image import Image
from kivy.graphics.transformation cimport Matrix

cdef class RenderContext(Canvas):
    '''The render context stores all the necessary information for drawing, i.e.:

    - The vertex shader
    - The fragment shader
    - The default texture
    - The state stack (color, texture, matrix...)
    '''
    def __cinit__(self, *args, **kwargs):
        self._use_parent_projection = 0
        self._use_parent_modelview = 0
        self._use_parent_frag_modelview = 0
        self.bind_texture = dict()

    def __init__(self, *args, **kwargs):
        Canvas.__init__(self, **kwargs)
        vs_src = kwargs.get('vs', None)
        fs_src = kwargs.get('fs', None)
        self._shader = Shader(vs_src, fs_src)

        # load default texture image
        filename = join(kivy_shader_dir, 'default.png')
        tex = Cache.get('kv.texture', filename)
        if not tex:
            tex = Image(filename).texture
            Cache.append('kv.texture', filename, tex)
        self.default_texture = tex

        self.state_stacks = {
            'opacity': [1.0],
            'texture0' : [0],
            'color'    : [[1.0,1.0,1.0,1.0]],
            'projection_mat': [Matrix()],
            'modelview_mat' : [Matrix()],
            'frag_modelview_mat' : [Matrix()],
        }

        cdef str key
        self._shader.use()
        for key, stack in self.state_stacks.iteritems():
            self.set_state(key, stack[0])

        if 'use_parent_projection' in kwargs:
            self._use_parent_projection = bool(int(kwargs['use_parent_projection']))
        if 'use_parent_modelview' in kwargs:
            self._use_parent_modelview = bool(int(kwargs['use_parent_modelview']))
        if 'use_parent_frag_modelview' in kwargs:
            self._use_parent_frag_modelview = bool(int(kwargs['use_parent_frag_modelview']))

    cdef void set_state(self, str name, value, int apply_now=0):
        # Upload the uniform value to the shader
        cdef list d
        if name not in self.state_stacks:
            self.state_stacks[name] = [value]
            self.flag_update()
        else:
            d = self.state_stacks[name]
            if value != d[-1]:
                d[-1] = value
                self.flag_update()
        self._shader.set_uniform(name, value)

    cdef get_state(self, str name):
        return self.state_stacks[name][-1]

    cdef int set_states(self, dict states) except -1:
        cdef str name
        for name, value in states.iteritems():
            self.set_state(name, value)

    cdef int push_state(self, str name) except -1:
        stack = self.state_stacks[name]
        stack.append(stack[-1])
        self.flag_update()

    cdef int push_states(self, list names) except -1:
        cdef str name
        for name in names:
            self.push_state(name)

    cdef int pop_state(self, str name) except -1:
        stack = self.state_stacks[name]
        oldvalue = stack.pop()
        if oldvalue != stack[-1]:
            self.set_state(name, stack[-1])
            self.flag_update()

    cdef int pop_states(self, list names) except -1:
        cdef str name
        for name in names:
            self.pop_state(name)

    cdef void set_texture(self, int index, Texture texture):
        # TODO this code is actually broken,
        # the binded texture can be already set, but we may changed if we came
        # from another render context.
        #if index in self.bind_texture and \
        #   self.bind_texture[index] is texture:
        #    return
        global _active_texture
        self.bind_texture[index] = texture
        if _active_texture != index:
            _active_texture = index
            cgl.glActiveTexture(GL_TEXTURE0 + index)
        texture.bind()
        self.flag_update()

    cdef int enter(self) except -1:
        self._shader.use()
        return 0

    cdef int leave(self) except -1:
        self._shader.stop()
        return 0

    cdef int apply(self) except -1:
        cdef list keys
        if PY2:
            keys = self.state_stacks.keys()
        else:
            keys = list(self.state_stacks.keys())

        cdef RenderContext active_context = getActiveContext()
        if self._use_parent_projection:
            self.set_state('projection_mat',
                    active_context.get_state('projection_mat'), 0)
        if self._use_parent_modelview:
            self.set_state('modelview_mat',
                    active_context.get_state('modelview_mat'), 0)
        if self._use_parent_frag_modelview:
            self.set_state('frag_modelview_mat',
                    active_context.get_state('frag_modelview_mat'), 0)
        pushActiveContext(self)
        if _need_reset_gl:
            reset_gl_context()
        self.push_states(keys)
        Canvas.apply(self)
        self.pop_states(keys)
        popActiveContext()
        self.flag_update_done()

        return 0

    cdef void reload(self) except *:
        pushActiveContext(self)
        reset_gl_context()
        Canvas.reload(self)
        popActiveContext()

    def __setitem__(self, key, val):
        self.set_state(key, val)

    def __getitem__(self, key):
        return self._shader.uniform_values[key]

    @property
    def shader(self):
        '''Return the shader attached to the render context.
        '''
        return self._shader

    @property
    def use_parent_projection(self):
        '''If True, the parent projection matrix will be used.

        .. versionadded:: 1.7.0

        Before::

            rc['projection_mat'] = Window.render_context['projection_mat']

        Now::

            rc = RenderContext(use_parent_projection=True)
        '''
        return bool(self._use_parent_projection)

    @use_parent_projection.setter
    def use_parent_projection(self, value):
        cdef cvalue = int(bool(value))
        if self._use_parent_projection != cvalue:
            self._use_parent_projection = cvalue
            self.flag_data_update()

    @property
    def use_parent_modelview(self):
        '''If True, the parent modelview matrix will be used.

        .. versionadded:: 1.7.0

        Before::

            rc['modelview_mat'] = Window.render_context['modelview_mat']

        Now::

            rc = RenderContext(use_parent_modelview=True)
        '''
        return bool(self._use_parent_modelview)

    @use_parent_modelview.setter
    def use_parent_modelview(self, value):
        cdef cvalue = int(bool(value))
        if self._use_parent_modelview != cvalue:
            self._use_parent_modelview = cvalue
            self.flag_data_update()

    @property
    def use_parent_frag_modelview(self):
        '''If True, the parent fragment modelview matrix will be used.

        .. versionadded:: 1.10.1

            rc = RenderContext(use_parent_frag_modelview=True)
        '''
        return bool(self._use_parent_frag_modelview)

    @use_parent_frag_modelview.setter
    def use_parent_frag_modelview(self, value):
        cdef cvalue = int(bool(value))
        if self._use_parent_frag_modelview != cvalue:
            self._use_parent_frag_modelview = cvalue
            self.flag_data_update()


cdef RenderContext ACTIVE_CONTEXT = None
cdef list CONTEXT_STACK  = list()


cdef RenderContext getActiveContext():
    global ACTIVE_CONTEXT
    return ACTIVE_CONTEXT


cdef pushActiveContext(RenderContext c):
    global CONTEXT_STACK, ACTIVE_CONTEXT
    CONTEXT_STACK.append(ACTIVE_CONTEXT)
    ACTIVE_CONTEXT = c
    c.enter()


cdef popActiveContext():
    global CONTEXT_STACK, ACTIVE_CONTEXT
    if ACTIVE_CONTEXT:
        ACTIVE_CONTEXT.leave()
    ACTIVE_CONTEXT = CONTEXT_STACK.pop()
    if ACTIVE_CONTEXT:
        ACTIVE_CONTEXT.enter()



'''
AVFoundation Camera
===================

Camera implementation using AVFoundation framework for OSX / iOS
'''

__all__ = ['CameraAVFoundation']

cdef extern from "camera_avfoundation_implem.h":
    ctypedef void *camera_t
    camera_t avf_camera_init(int index, int width, int height)
    void avf_camera_deinit(camera_t camera)
    bint avf_camera_update(camera_t camera)
    void avf_camera_start(camera_t camera)
    void avf_camera_stop(camera_t camera)
    void avf_camera_get_image(camera_t camera, int *width, int *height, int *rowsize, char **data)
    bint avf_camera_attempt_framerate_selection(camera_t camera, int fps)
    bint avf_camera_attempt_capture_preset(camera_t camera, char* preset)
    bint avf_camera_attempt_start_metadata_analysis(camera_t camera)
    void avf_camera_get_metadata(camera_t camera, char **metatype, char **data)
    bint avf_camera_have_new_metadata(camera_t camera);
    bint avf_camera_set_video_orientation(camera_t camera, int orientation)


from kivy.logger import Logger
from kivy.clock import Clock
from kivy.graphics.texture import Texture
from kivy.core.camera import CameraBase
from kivy.utils import platform
from cython cimport view as cyview


cdef class _AVStorage:
    cdef camera_t camera

    def __cinit__(self):
        self.camera = NULL


class CameraAVFoundation(CameraBase):
    '''Implementation of CameraBase using AVFoundation
    '''

    def __init__(self, **kwargs):
        self._storage = _AVStorage()
        self._update_ev = None
        self._metadata_callback = None
        self._framerate = 30
        super(CameraAVFoundation, self).__init__(**kwargs)

    def __del__(self):
        self._release_camera()

    def init_camera(self):
        cdef _AVStorage storage = <_AVStorage>self._storage
        storage.camera = avf_camera_init(
            self._index, self.resolution[0], self.resolution[1])

    def _release_camera(self):
        cdef _AVStorage storage = <_AVStorage>self._storage
        if storage.camera != NULL:
            avf_camera_deinit(storage.camera)
            storage.camera = NULL

    @property
    def _scheduled_rate(self):
        # We're going 4 times faster the framerate to avoid frame skipping.
        return 1 / (self._framerate * 4)

    def _update(self, dt):
        cdef _AVStorage storage = <_AVStorage>self._storage
        cdef int width, height, rowsize
        cdef char *data
        cdef char *metadata_type
        cdef char *metadata_data
        cdef cyview.array cyarr

        if self.stopped:
            return

        if not avf_camera_update(storage.camera):
            return

        avf_camera_get_image(storage.camera,
            &width, &height, &rowsize, &data)

        if data == NULL:
            return

        cyarr = cyview.array(
            shape=(rowsize * height,),
            itemsize=sizeof(char),
            format="B",
            mode="c",
            allocate_buffer=False,
        )
        cyarr.data = data

        self._resolution = (width, height)
        
        if self._texture is None or self._texture.size != self._resolution:
            if platform == 'ios':
                self._texture = Texture.create(self._resolution, colorfmt='bgra')
            else:
                self._texture = Texture.create(self._resolution)
            self._texture.flip_vertical()
            self.dispatch('on_load')

        self._format = 'bgra'
        self._texture.blit_buffer(cyarr, colorfmt=self._format)
        self._copy_to_gpu()
        if self._metadata_callback:
            if avf_camera_have_new_metadata(storage.camera):
                avf_camera_get_metadata(storage.camera, &metadata_type, &metadata_data)
                self._metadata_callback(metadata_type, metadata_data)
    
    def _copy_to_gpu(self):
        self.dispatch('on_texture')

    def start(self):
        cdef _AVStorage storage = <_AVStorage>self._storage
        super(CameraAVFoundation, self).start()
        if self._update_ev is not None:
            self._update_ev.cancel()
        self._update_ev = Clock.schedule_interval(self._update, self._scheduled_rate)
        avf_camera_start(storage.camera)

    def stop(self):
        cdef _AVStorage storage = <_AVStorage>self._storage
        super(CameraAVFoundation, self).stop()
        if self._update_ev is not None:
            self._update_ev.cancel()
            self._update_ev = None
        avf_camera_stop(storage.camera)

    def set_framerate(self, framerate):
        cdef _AVStorage storage = <_AVStorage>self._storage
        if avf_camera_attempt_framerate_selection(storage.camera, framerate):
            if self._update_ev is not None:
                self._update_ev.cancel()
            self._framerate = framerate
            self._update_ev = Clock.schedule_interval(self._update, self._scheduled_rate)

    def set_preset(self, preset):
        cdef _AVStorage storage = <_AVStorage>self._storage
        avf_camera_attempt_capture_preset(storage.camera, preset)

    def set_video_orientation(self, orientation):
        cdef _AVStorage storage = <_AVStorage>self._storage
        avf_camera_set_video_orientation(storage.camera, orientation)

    def start_metadata_analysis(self, callback=None):
        cdef _AVStorage storage = <_AVStorage>self._storage
        self._metadata_callback = callback
        avf_camera_attempt_start_metadata_analysis(storage.camera)


# -*- coding: utf-8 -*-
from sphinx.ext.autodoc import Documenter, ClassDocumenter
from sphinx.ext.autodoc import setup as core_setup
from sphinx.locale import _


class KivyClassDocumenter(ClassDocumenter):
    def add_directive_header(self, sig):
        if self.doc_as_attr:
            self.directivetype = 'attribute'
        Documenter.add_directive_header(self, sig)

        def fix(mod):
            if mod == 'kivy._event':
                mod = 'kivy.event'
            return mod

        # add inheritance info, if wanted
        if not self.doc_as_attr and self.options.show_inheritance:
            self.add_line('', '<autodoc>')
            if len(self.object.__bases__):
                bases = [b.__module__ == '__builtin__' and
                         ':class:`%s`' % b.__name__ or
                         ':class:`%s.%s`' % (fix(b.__module__), b.__name__)
                         for b in self.object.__bases__]
                self.add_line(_('   Bases: %s') % ', '.join(bases),
                              '<autodoc>')
def setup(app):
    core_setup(app)
    app.add_autodocumenter(KivyClassDocumenter)



'''
Vertex Buffer
=============

The :class:`VBO` class handles the creation and updating of Vertex Buffer
Objects in OpenGL.

.. versionadded:: 1.6.0
    VertexFormat class added. VertexFormat is used to describe the layout
    of the vertex data stored in vertex arrays/vbo's. The default vertex format
    is:
        VertexFormat(('vPosition', 2, 'float'), ('vTexCoords0', 2, 'float'))

.. versionchanged:: 1.6.0
    VBO now no longer has a fixed vertex format. If no VertexFormat is given
    at initialization, the default vertex format is used.
'''

__all__ = ('VBO', 'VertexBatch', 'VertexFormat')

include "../include/config.pxi"
include "common.pxi"
include "gl_debug_logger.pxi"

from os import environ
from kivy.graphics.buffer cimport Buffer
from kivy.graphics.cgl cimport *
from kivy.graphics.vertex cimport *
from kivy.logger import Logger
from kivy.graphics.context cimport Context, get_context
from kivy.graphics.instructions cimport getActiveContext
from kivy.graphics.shader cimport Shader

cdef VertexFormat default_vertex = VertexFormat( (b'vPosition', 2, 'float'),
        (b'vTexCoords0', 2, 'float'))

cdef short V_NEEDGEN = 1 << 0
cdef short V_NEEDUPLOAD = 1 << 1
cdef short V_HAVEID = 1 << 2

cdef class VBO:
    '''
    .. versionchanged:: 1.6.0
        VBO now no longer has a fixed vertex format. If no VertexFormat is given
        at initialization, the default vertex format is used.
    '''
    def __cinit__(self, VertexFormat vertex_format=None):
        self.usage  = GL_DYNAMIC_DRAW
        self.target = GL_ARRAY_BUFFER
        if vertex_format is None:
            vertex_format = default_vertex
        self.vertex_format = vertex_format
        self.format = vertex_format.vattr
        self.format_count = vertex_format.vattr_count
        self.format_size = vertex_format.vbytesize
        self.flags = V_NEEDGEN | V_NEEDUPLOAD
        self.vbo_size = 0

    def __dealloc__(self):
        get_context().dealloc_vbo(self)

    def __init__(self, VertexFormat vertex_format=None):
        self.data = Buffer(self.format_size)

    cdef int have_id(self):
        return self.flags & V_HAVEID

    cdef void update_buffer(self):
        # generate VBO if not done yet
        if self.flags & V_NEEDGEN:
            cgl.glGenBuffers(1, &self.id)
            log_gl_error('VBO.update_buffer-glGenBuffers')
            self.flags &= ~V_NEEDGEN
            self.flags |= V_HAVEID

        # if the size doesn't match, we need to reupload the whole data
        if self.vbo_size < self.data.size():
            self.vbo_size = self.data.size()
            cgl.glBindBuffer(GL_ARRAY_BUFFER, self.id)
            log_gl_error('VBO.update_buffer-glBindBuffer')
            cgl.glBufferData(GL_ARRAY_BUFFER, self.vbo_size, self.data.pointer(),
                self.usage)
            log_gl_error('VBO.update_buffer-glBufferData')
            self.flags &= ~V_NEEDUPLOAD

        # if size match, update only what is needed
        elif self.flags & V_NEEDUPLOAD:
            cgl.glBindBuffer(GL_ARRAY_BUFFER, self.id)
            log_gl_error('VBO.update_buffer-glBindBuffer')
            cgl.glBufferSubData(GL_ARRAY_BUFFER, 0, self.data.size(),
                self.data.pointer())
            log_gl_error('VBO.update_buffer-glBufferSubData')
            self.flags &= ~V_NEEDUPLOAD

    cdef void bind(self):
        cdef Shader shader = getActiveContext()._shader
        cdef vertex_attr_t *attr
        cdef int offset = 0, i
        self.update_buffer()
        cgl.glBindBuffer(GL_ARRAY_BUFFER, self.id)
        log_gl_error('VBO.bind-glBindBuffer')
        shader.bind_vertex_format(self.vertex_format)
        for i in xrange(self.format_count):
            attr = &self.format[i]
            if attr.per_vertex == 0:
                continue
            if attr.index != <unsigned int>-1:
                cgl.glVertexAttribPointer(attr.index, attr.size, attr.type,
                        GL_FALSE, <GLsizei>self.format_size, <GLvoid*><unsigned int>offset)
                log_gl_error('VBO.bind-glVertexAttribPointer')
            offset += attr.bytesize

    cdef void unbind(self):
        cgl.glBindBuffer(GL_ARRAY_BUFFER, 0)
        log_gl_error('VBO.unbind-glBindBuffer')

    cdef void add_vertex_data(self, void *v, unsigned short* indices,
        int count):
        self.flags |= V_NEEDUPLOAD
        self.data.add(v, indices, count)

    cdef void update_vertex_data(self, int index, void* v, int count):
        self.flags |= V_NEEDUPLOAD
        self.data.update(index, v, count)

    cdef void remove_vertex_data(self, unsigned short* indices, int count):
        self.data.remove(indices, count)

    cdef void reload(self):
        self.flags = V_NEEDUPLOAD | V_NEEDGEN
        self.vbo_size = 0

    def __repr__(self):
        return '<VBO at %x id=%r count=%d size=%d>' % (
                id(self), self.id if self.flags & V_HAVEID else None,
                self.data.count(), self.data.size())

cdef class VertexBatch:
    def __init__(self, **kwargs):
        self.usage  = GL_DYNAMIC_DRAW
        cdef object lushort = sizeof(unsigned short)
        self.vbo = kwargs.get('vbo')
        if self.vbo is None:
            self.vbo = VBO()
        self.vbo_index = Buffer(lushort) #index of every vertex in the vbo
        self.elements = Buffer(lushort) #indices translated to vbo indices
        self.elements_size = 0
        self.flags = V_NEEDGEN | V_NEEDUPLOAD

        self.set_data(NULL, 0, NULL, 0)
        self.set_mode(kwargs.get('mode'))

    def __dealloc__(self):
        get_context().dealloc_vertexbatch(self)

    cdef int have_id(self):
        return self.flags & V_HAVEID

    cdef void reload(self):
        self.flags = V_NEEDGEN | V_NEEDUPLOAD
        self.elements_size = 0

    cdef void clear_data(self):
        # clear old vertices from vbo and then reset index buffer
        self.vbo.remove_vertex_data(<unsigned short*>self.vbo_index.pointer(),
                                    self.vbo_index.count())
        self.vbo_index.clear()
        self.elements.clear()

    cdef void set_data(self, void *vertices, int vertices_count,
                       unsigned short *indices, int indices_count):
        #clear old vertices first
        self.clear_data()
        self.elements.grow(indices_count)

        # now append the vertices and indices to vbo
        #vsize = self.vbo.vertex_format.vsize
        self.append_data(vertices, vertices_count, indices, indices_count)
        self.flags |= V_NEEDUPLOAD

    cdef void append_data(self, void *vertices, int vertices_count,
                          unsigned short *indices, int indices_count):
        # add vertex data to vbo and get index for every vertex added
        cdef unsigned short *vi = <unsigned short *>malloc(sizeof(
            unsigned short) * vertices_count)
        if vi == NULL:
            raise MemoryError('vertex index allocation')
        self.vbo.add_vertex_data(vertices, vi, vertices_count)
        self.vbo_index.add(vi, NULL, vertices_count)
        free(vi)

        # build element list for DrawElements using vbo indices
        # TODO: remove buffer usage in this case, the memory is always one big
        # block. no need to use add() everytime we need to reconstruct the list.
        cdef int local_index
        cdef unsigned short *vbi = <unsigned short*>self.vbo_index.pointer()
        for i in xrange(indices_count):
            local_index = indices[i]
            self.elements.add(&vbi[local_index], NULL, 1)
        self.flags |= V_NEEDUPLOAD

    cdef void draw(self):
        cdef int count = self.elements.count()
        if count == 0:
            return

        # create when needed
        if self.flags & V_NEEDGEN:
            cgl.glGenBuffers(1, &self.id)
            self.flags &= ~V_NEEDGEN
            self.flags |= V_HAVEID

        # bind to the current id
        cgl.glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, self.id)
        log_gl_error('VertexBatch.draw-glBindBuffer')

        # cache indices in a gpu buffer too
        if self.flags & V_NEEDUPLOAD:
            if self.elements_size == self.elements.size():
                cgl.glBufferSubData(GL_ELEMENT_ARRAY_BUFFER, 0, self.elements_size,
                    self.elements.pointer())
                log_gl_error('VertexBatch.draw-glBufferSubData')
            else:
                cgl.glBufferData(GL_ELEMENT_ARRAY_BUFFER, self.elements.size(),
                    self.elements.pointer(), self.usage)
                self.elements_size = self.elements.size()
                log_gl_error('VertexBatch.draw-glBufferData')
            self.flags &= ~V_NEEDUPLOAD

        self.vbo.bind()

        # draw the elements pointed by indices in ELEMENT ARRAY BUFFER.
        cgl.glDrawElements(self.mode, count, GL_UNSIGNED_SHORT, NULL)
        log_gl_error('VertexBatch.draw-glDrawElements')

    cdef void set_mode(self, str mode):
        # most common case in top;
        self.mode_str = mode
        if mode is None:
            self.mode = GL_TRIANGLES
        elif mode == 'points':
            self.mode = GL_POINTS
        elif mode == 'line_strip':
            self.mode = GL_LINE_STRIP
        elif mode == 'line_loop':
            self.mode = GL_LINE_LOOP
        elif mode == 'lines':
            self.mode = GL_LINES
        elif mode == 'triangle_strip':
            self.mode = GL_TRIANGLE_STRIP
        elif mode == 'triangle_fan':
            self.mode = GL_TRIANGLE_FAN
        else:
            self.mode = GL_TRIANGLES

    cdef str get_mode(self):
        return self.mode_str

    cdef int count(self):
        return self.elements.count()

    def __repr__(self):
        return '<VertexBatch at %x id=%r vertex=%d size=%d mode=%s vbo=%x>' % (
                id(self), self.id if self.flags & V_HAVEID else None,
                self.elements.count(), self.elements.size(), self.get_mode(),
                id(self.vbo))







cdef class Rect:
    def __cinit__(self, int32_t x, int32_t y, int32_t width, int32_t height):
        self._vc_rect.x = x
        self._vc_rect.y = y
        self._vc_rect.width = width
        self._vc_rect.height = height
        
    @property
    def x(self):
        return self._vc_rect.x
            
    @x.setter
    def x(self, int32_t x):
        self._vc_rect.x = x
            
    @property
    def y(self):
        return self._vc_rect.y
            
    @y.setter
    def y(self, int32_t y):
        self._vc_rect.y = y
            
    @property
    def width(self):
        return self._vc_rect.width
            
    @width.setter
    def width(self, int32_t width):
        self._vc_rect.width = width
            
    @property
    def height(self):
        return self._vc_rect.height
            
    @height.setter
    def height(self, int32_t h):
        self._vc_rect.height = h
            
            
cdef class DisplayHandle:
    pass
    
cdef class UpdateHandle:
    pass
    
cdef class ResourceHandle:
    pass
    
cdef class ProtectionHandle:
    pass
    
cdef class ElementHandle:
    pass
    
    
class BCMDisplayException(Exception):
    pass
    
    
def host_init():
    bcm_host_init()
    
def host_deinit():
    bcm_host_deinit()
    
def graphics_get_display_size(uint16_t number):
    cdef:
        int32_t ret
        uint32_t width, height
        
    ret = c_get_display_size(number, &width, &height)
    if ret < 0:
        raise BCMDisplayException("Failed to get display size")
    return (width, height)

def display_open(uint32_t device):
    cdef:
        DISPMANX_DISPLAY_HANDLE_T disp
        DisplayHandle D
    disp = vc_dispmanx_display_open( 0 )
    print('Apenned display handle:', disp)
    if disp == 0:
        raise BCMDisplayException("Couldn't open handle to display")
    D = DisplayHandle()
    D._handle = disp
    return D
    
def update_start(int32_t priority):
    cdef DISPMANX_UPDATE_HANDLE_T hdl
    hdl = vc_dispmanx_update_start( priority )
    U = UpdateHandle()
    U._handle = hdl
    return U
    
def element_add(UpdateHandle update, 
                DisplayHandle display, 
                int32_t layer,
                Rect dest_rect,
                Rect src_rect):
    cdef:
        DISPMANX_ELEMENT_HANDLE_T elem
    elem = vc_dispmanx_element_add (update._handle, 
                                     display._handle,
                                     layer, 
                                     &(dest_rect._vc_rect), 
                                     0, #DISPMANX_RESOURCE_HANDLE_T src,
                                     &(src_rect._vc_rect), 
                                     0, #DISPMANX_PROTECTION_T protection, 
                                     <VC_DISPMANX_ALPHA_T *>0, #VC_DISPMANX_ALPHA_T *alpha,
                                     <DISPMANX_CLAMP_T *>0, #DISPMANX_CLAMP_T *clamp, 
                                     <DISPMANX_TRANSFORM_T>0) #DISPMANX_TRANSFORM_T transform
    E = ElementHandle()
    E._handle = elem
    return E
    
def update_submit_sync(UpdateHandle update):
    return vc_dispmanx_update_submit_sync( update._handle )



'''
Texture
=======

.. versionchanged:: 1.6.0
    Added support for paletted texture on OES: 'palette4_rgb8',
    'palette4_rgba8', 'palette4_r5_g6_b5', 'palette4_rgba4', 'palette4_rgb5_a1',
    'palette8_rgb8', 'palette8_rgba8', 'palette8_r5_g6_b5', 'palette8_rgba4'
    and 'palette8_rgb5_a1'.

:class:`Texture` is a class that handles OpenGL textures. Depending on the
hardware,
some OpenGL capabilities might not be available (BGRA support, NPOT support,
etc.)

You cannot instantiate this class yourself. You must use the function
:meth:`Texture.create` to create a new texture::

    texture = Texture.create(size=(640, 480))

When you create a texture, you should be aware of the default color
and buffer format:

    - the color/pixel format (:attr:`Texture.colorfmt`) that can be one of
      'rgb', 'rgba', 'luminance', 'luminance_alpha', 'bgr' or 'bgra'.
      The default value is 'rgb'
    - the buffer format determines how a color component is stored into memory.
      This can be one of 'ubyte', 'ushort', 'uint', 'byte', 'short', 'int' or
      'float'. The default value and the most commonly used is 'ubyte'.

So, if you want to create an RGBA texture::

    texture = Texture.create(size=(640, 480), colorfmt='rgba')

You can use your texture in almost all vertex instructions with the
:attr:`kivy.graphics.VertexIntruction.texture` parameter. If you want to use
your texture in kv lang, you can save it in an
:class:`~kivy.properties.ObjectProperty` inside your widget.

.. warning::
    Using Texture before OpenGL has been initialized will lead to a crash. If
    you need to create textures before the application has started, import
    Window first: `from kivy.core.window import Window`

Blitting custom data
--------------------

You can create your own data and blit it to the texture using
:meth:`Texture.blit_buffer`.

For example, to blit immutable bytes data::

    # create a 64x64 texture, defaults to rgba / ubyte
    texture = Texture.create(size=(64, 64))

    # create 64x64 rgb tab, and fill with values from 0 to 255
    # we'll have a gradient from black to white
    size = 64 * 64 * 3
    buf = [int(x * 255 / size) for x in range(size)]

    # then, convert the array to a ubyte string
    buf = bytes(buf)

    # then blit the buffer
    texture.blit_buffer(buf, colorfmt='rgb', bufferfmt='ubyte')

    # that's all ! you can use it in your graphics now :)
    # if self is a widget, you can do this
    with self.canvas:
        Rectangle(texture=texture, pos=self.pos, size=(64, 64))

Since 1.9.0, you can blit data stored in a instance that implements the python
buffer interface, or a memoryview thereof, such as numpy arrays, python
`array.array`, a `bytearray`, or a cython array. This is beneficial if you
expect to blit similar data, with perhaps a few changes in the data.

When using a bytes representation of the data, for every change you have to
regenerate the bytes instance, from perhaps a list, which is very inefficient.
When using a buffer object, you can simply edit parts of the original data.
Similarly, unless starting with a bytes object, converting to bytes requires a
full copy, however, when using a buffer instance, no memory is copied, except
to upload it to the GPU.

Continuing with the example above::

    from array import array

    size = 64 * 64 * 3
    buf = [int(x * 255 / size) for x in range(size)]
    # initialize the array with the buffer values
    arr = array('B', buf)
    # now blit the array
    texture.blit_buffer(arr, colorfmt='rgb', bufferfmt='ubyte')

    # now change some elements in the original array
    arr[24] = arr[50] = 99
    # blit again the buffer
    texture.blit_buffer(arr, colorfmt='rgb', bufferfmt='ubyte')


BGR/BGRA support
----------------

The first time you try to create a BGR or BGRA texture, we check whether
your hardware supports BGR / BGRA textures by checking the extension
'GL_EXT_bgra'.

If the extension is not found, the conversion to RGB / RGBA will be done in
software.


NPOT texture
------------

.. versionchanged:: 1.0.7

    If your hardware supports NPOT, no POT is created.

As the OpenGL documentation says, a texture must be power-of-two sized. That
means
your width and height can be one of 64, 32, 256... but not 3, 68, 42. NPOT means
non-power-of-two. OpenGL ES 2 supports NPOT textures natively but with some
drawbacks. Another type of NPOT texture is called a rectangle texture.
POT, NPOT and textures all have their own pro/cons.

================= ============= ============= =================================
    Features           POT           NPOT                Rectangle
----------------- ------------- ------------- ---------------------------------
OpenGL Target     GL_TEXTURE_2D GL_TEXTURE_2D GL_TEXTURE_RECTANGLE_(NV|ARB|EXT)
Texture coords    0-1 range     0-1 range     width-height range
Mipmapping        Supported     Partially     No
Wrap mode         Supported     Supported     No
================= ============= ============= =================================

If you create a NPOT texture, we first check whether your hardware
supports it by checking the extensions GL_ARB_texture_non_power_of_two or
OES_texture_npot. If none of these are available, we create the nearest
POT texture that can contain your NPOT texture. The :meth:`Texture.create` will
return a :class:`TextureRegion` instead.


Texture atlas
-------------

A texture atlas is a single texture that contains many images.
If you want to separate the original texture into many single ones, you don't
need to. You can get a region of the original texture. That will return the
original texture with custom texture coordinates::

    # for example, load a 128x128 image that contain 4 64x64 images
    from kivy.core.image import Image
    texture = Image('mycombinedimage.png').texture

    bottomleft = texture.get_region(0, 0, 64, 64)
    bottomright = texture.get_region(0, 64, 64, 64)
    topleft = texture.get_region(0, 64, 64, 64)
    topright = texture.get_region(64, 64, 64, 64)


.. _mipmap:

Mipmapping
----------

.. versionadded:: 1.0.7

Mipmapping is an OpenGL technique for enhancing the rendering of large textures
to small surfaces. Without mipmapping, you might see pixelation when you
render to small surfaces.
The idea is to precalculate the subtexture and apply some image filter as a
linear filter. Then, when you render a small surface, instead of using the
biggest texture, it will use a lower filtered texture. The result can look
better this way.

To make that happen, you need to specify mipmap=True when you create a
texture. Some widgets already give you the ability to create mipmapped
textures, such as the :class:`~kivy.uix.label.Label` and
:class:`~kivy.uix.image.Image`.

From the OpenGL Wiki : "So a 64x16 2D texture can have 5 mip-maps: 32x8, 16x4,
8x2, 4x1, 2x1, and 1x1". Check http://www.opengl.org/wiki/Texture for more
information.

.. note::

    As the table in previous section said, if your texture is NPOT, we
    create the nearest POT texture and generate a mipmap from it. This
    might change in the future.

Reloading the Texture
---------------------

.. versionadded:: 1.2.0

If the OpenGL context is lost, the Texture must be reloaded. Textures that have
a source are automatically reloaded but generated textures must
be reloaded by the user.

Use the :meth:`Texture.add_reload_observer` to add a reloading function that
will be automatically called when needed::

    def __init__(self, **kwargs):
        super(...).__init__(**kwargs)
        self.texture = Texture.create(size=(512, 512), colorfmt='RGB',
            bufferfmt='ubyte')
        self.texture.add_reload_observer(self.populate_texture)

        # and load the data now.
        self.cbuffer = '\\x00\\xf0\\xff' * 512 * 512
        self.populate_texture(self.texture)

    def populate_texture(self, texture):
        texture.blit_buffer(self.cbuffer)

This way, you can use the same method for initialization and reloading.

.. note::

    For all text rendering with our core text renderer, the texture is generated
    but we already bind a method to redo the text rendering and reupload
    the text to the texture. You don't have to do anything.
'''

__all__ = ('Texture', 'TextureRegion')

include "../include/config.pxi"
include "common.pxi"
include "opengl_utils_def.pxi"
include "img_tools.pxi"
include "gl_debug_logger.pxi"

cimport cython
from os import environ
from kivy.utils import platform
from kivy.weakmethod import WeakMethod
from kivy.graphics.context cimport get_context

cimport kivy.graphics.cgl as cgldef
from kivy.graphics.cgl cimport *
from kivy.graphics.opengl_utils cimport gl_has_capability, gl_get_version_major

cdef int gles_limts = int(environ.get(
    'KIVY_GLES_LIMITS', int(platform not in ('win', 'macosx', 'linux'))))

# update flags
cdef int TI_MIN_FILTER      = 1 << 0
cdef int TI_MAG_FILTER      = 1 << 1
cdef int TI_WRAP            = 1 << 2
cdef int TI_NEED_GEN        = 1 << 3
cdef int TI_NEED_ALLOCATE   = 1 << 4
cdef int TI_NEED_PIXELS     = 1 << 5

# compatibility layer
DEF GL_BGR = 0x80E0
DEF GL_BGRA = 0x80E1
DEF GL_COMPRESSED_RGBA_S3TC_DXT1_EXT = 0x83F1
DEF GL_COMPRESSED_RGBA_S3TC_DXT3_EXT = 0x83F2
DEF GL_COMPRESSED_RGBA_S3TC_DXT5_EXT = 0x83F3
DEF GL_ETC1_RGB8_OES = 0x8D64
DEF GL_PALETTE4_RGB8_OES = 0x8B90
DEF GL_PALETTE4_RGBA8_OES = 0x8B91
DEF GL_PALETTE4_R5_G6_B5_OES = 0x8B92
DEF GL_PALETTE4_RGBA4_OES = 0x8B93
DEF GL_PALETTE4_RGB5_A1_OES = 0x8B94
DEF GL_PALETTE8_RGB8_OES = 0x8B95
DEF GL_PALETTE8_RGBA8_OES = 0x8B96
DEF GL_PALETTE8_R5_G6_B5_OES = 0x8B97
DEF GL_PALETTE8_RGBA4_OES = 0x8B98
DEF GL_PALETTE8_RGB5_A1_OES = 0x8B99
DEF GL_COMPRESSED_RGB_PVRTC_4BPPV1_IMG = 0x8C00
DEF GL_COMPRESSED_RGB_PVRTC_2BPPV1_IMG = 0x8C01
DEF GL_COMPRESSED_RGBA_PVRTC_4BPPV1_IMG = 0x8C02
DEF GL_COMPRESSED_RGBA_PVRTC_2BPPV1_IMG = 0x8C03
DEF GL_RED = 0x1903
DEF GL_RG = 0x8227
DEF GL_R8 = 0x8229
DEF GL_RG8 = 0x822B
DEF GL_RGBA8 =  0x8058
DEF GL_UNPACK_ROW_LENGTH = 0x0CF2
DEF GL_UNPACK_SKIP_ROWS = 0x0CF3
DEF GL_UNPACK_SKIP_PIXELS = 0x0CF4

cdef dict _gl_color_fmt = {
    'rgba': GL_RGBA, 'bgra': GL_BGRA, 'rgb': GL_RGB, 'bgr': GL_BGR,
    'luminance': GL_LUMINANCE, 'luminance_alpha': GL_LUMINANCE_ALPHA,
    's3tc_dxt1': GL_COMPRESSED_RGBA_S3TC_DXT1_EXT,
    's3tc_dxt3': GL_COMPRESSED_RGBA_S3TC_DXT3_EXT,
    's3tc_dxt5': GL_COMPRESSED_RGBA_S3TC_DXT5_EXT,
    'etc1_rgb8': GL_ETC1_RGB8_OES,
    'palette4_rgb8': GL_PALETTE4_RGB8_OES,
    'palette4_rgba8': GL_PALETTE4_RGBA8_OES,
    'palette4_r5_g6_b5': GL_PALETTE4_R5_G6_B5_OES,
    'palette4_rgba4': GL_PALETTE4_RGBA4_OES,
    'palette4_rgb5_a1': GL_PALETTE4_RGB5_A1_OES,
    'palette8_rgb8': GL_PALETTE8_RGB8_OES,
    'palette8_rgba8': GL_PALETTE8_RGBA8_OES,
    'palette8_r5_g6_b5': GL_PALETTE8_R5_G6_B5_OES,
    'palette8_rgba4': GL_PALETTE8_RGBA4_OES,
    'palette8_rgb5_a1': GL_PALETTE8_RGB5_A1_OES,
    'pvrtc_rgba2': GL_COMPRESSED_RGBA_PVRTC_2BPPV1_IMG,
    'pvrtc_rgba4': GL_COMPRESSED_RGBA_PVRTC_4BPPV1_IMG,
    'pvrtc_rgb2': GL_COMPRESSED_RGB_PVRTC_2BPPV1_IMG,
    'pvrtc_rgb4': GL_COMPRESSED_RGB_PVRTC_4BPPV1_IMG,
    'red': GL_RED, 'rg': GL_RG,
    'r8': GL_R8, 'rg8': GL_RG8, 'rgba8': GL_RGBA8}

cdef dict _gl_buffer_fmt = {
    'ubyte': GL_UNSIGNED_BYTE, 'ushort': GL_UNSIGNED_SHORT,
    'uint': GL_UNSIGNED_INT, 'byte': GL_BYTE,
    'short': GL_SHORT, 'int': GL_INT, 'float': GL_FLOAT}


cdef dict _gl_buffer_size = {
    'ubyte': sizeof(GLubyte), 'ushort': sizeof(GLushort),
    'uint': sizeof(GLuint), 'byte': sizeof(GLbyte),
    'short': sizeof(GLshort), 'int': sizeof(GLint),
    'float': sizeof(GLfloat) }


cdef dict _gl_texture_min_filter = {
    'nearest': GL_NEAREST, 'linear': GL_LINEAR,
    'nearest_mipmap_nearest': GL_NEAREST_MIPMAP_NEAREST,
    'nearest_mipmap_linear': GL_NEAREST_MIPMAP_LINEAR,
    'linear_mipmap_nearest': GL_LINEAR_MIPMAP_NEAREST,
    'linear_mipmap_linear': GL_LINEAR_MIPMAP_LINEAR }


cdef inline int _nearest_pow2(int v):
    # From http://graphics.stanford.edu/~seander/bithacks.html#RoundUpPowerOf2
    # Credit: Sean Anderson
    v -= 1
    v |= v >> 1
    v |= v >> 2
    v |= v >> 4
    v |= v >> 8
    v |= v >> 16
    return v + 1


cdef inline int _is_pow2(int v):
    # http://graphics.stanford.edu/~seander/bithacks.html#DetermineIfPowerOf2
    return (v & (v - 1)) == 0


cdef inline int _color_fmt_to_gl(x):
    '''Return the GL numeric value from a color string format.
    '''
    x = x.lower()
    try:
        return _gl_color_fmt[x]
    except KeyError:
        raise Exception('Unknown <%s> color format' % x)


cdef inline int _is_compressed_fmt(x):
    '''Return 1 if the color string format is a compressed one.
    '''
    if x.startswith('palette'):
        return 1
    if x.startswith('pvrtc_'):
        return 1
    if x.startswith('etc1_'):
        return 1
    return x.startswith('s3tc_dxt')


cdef inline int _buffer_fmt_to_gl(x):
    '''Return the GL numeric value from a buffer string format.
    '''
    x = x.lower()
    try:
        return _gl_buffer_fmt[x]
    except KeyError:
        raise Exception('Unknown <%s> buffer format' % x)


cdef inline int _buffer_type_to_gl_size(x):
    '''Return the size of a buffer string format in str.
    '''
    x = x.lower()
    try:
        return _gl_buffer_size[x]
    except KeyError:
        raise Exception('Unknown <%s> format' % x)


cdef inline GLuint _str_to_gl_texture_min_filter(x):
    '''Return the GL numeric value from a texture min filter string.
    '''
    x = x.lower()
    try:
        return _gl_texture_min_filter[x]
    except KeyError:
        raise Exception('Unknown <%s> texture min filter' % x)


cdef inline GLuint _str_to_gl_texture_mag_filter(x):
    '''Return the GL numeric value from a texture mag filter string.
    '''
    x = x.lower()
    if x == 'nearest':
        return GL_NEAREST
    elif x == 'linear':
        return GL_LINEAR
    raise Exception('Unknown <%s> texture mag filter' % x)


cdef inline GLuint _str_to_gl_texture_wrap(x):
    '''Return the GL numeric value from a texture wrap string.
    '''
    if x == 'clamp_to_edge':
        return GL_CLAMP_TO_EDGE
    elif x == 'repeat':
        return GL_REPEAT
    elif x == 'mirrored_repeat':
        return GL_MIRRORED_REPEAT


cdef inline int _gl_format_size(GLuint x):
    '''Return the GL numeric value from a texture wrap string.
    '''
    if x in (GL_RGB, GL_BGR):
        return 3
    elif x in (GL_RGBA, GL_BGRA):
        return 4
    elif x in (GL_LUMINANCE_ALPHA, GL_RG):
        return 2
    elif x in (GL_LUMINANCE, GL_RED):
        return 1
    elif x in (GL_COMPRESSED_RGBA_S3TC_DXT1_EXT,
            GL_COMPRESSED_RGBA_S3TC_DXT3_EXT,
            GL_COMPRESSED_RGBA_S3TC_DXT5_EXT):
        return 4
    raise Exception('Unsupported format size <%s>' % str(format))


cdef inline int _is_gl_format_supported(x):
    if x in ('bgr', 'bgra'):
        return gl_has_capability(GLCAP_BGRA)
    elif x == 's3tc_dxt1':
        return gl_has_capability(GLCAP_DXT1)
    elif x.startswith('s3tc_dxt'):
        return gl_has_capability(GLCAP_S3TC)
    elif x.startswith('etc1_'):
        return gl_has_capability(GLCAP_ETC1)
    return 1


cdef inline str _convert_gl_format(x):
    if x == 'bgr':
        return 'rgb'
    elif x == 'bgra':
        return 'rgba'
    return x


cdef inline void _gl_prepare_pixels_upload(int width) nogil:
    '''Set the best pixel alignment for the current width.
    '''
    if not (width & 0x7):
        cgl.glPixelStorei(GL_UNPACK_ALIGNMENT, 8)
    elif not (width & 0x3):
        cgl.glPixelStorei(GL_UNPACK_ALIGNMENT, 4)
    elif not (width & 0x1):
        cgl.glPixelStorei(GL_UNPACK_ALIGNMENT, 2)
    else:
        cgl.glPixelStorei(GL_UNPACK_ALIGNMENT, 1)



cdef Texture _texture_create(int width, int height, colorfmt, bufferfmt,
                     int mipmap, int allocate, object callback, object icolorfmt):
    '''Create the OpenGL texture.
    '''
    cdef GLuint target = GL_TEXTURE_2D
    cdef GLuint texid = 0
    cdef Texture texture
    cdef int texture_width, texture_height
    cdef int glbufferfmt = _buffer_fmt_to_gl(bufferfmt)
    cdef int make_npot = 0

    # check if it's a pot or not
    if not _is_pow2(width) or not _is_pow2(height):
        make_npot = 1

    if not cgldef.kivy_opengl_es2:
        if gl_get_version_major() < 3:
            mipmap = 0

    # in case of mipmap is asked for npot texture, make it pot compatible
    if mipmap:
        make_npot = 0
        allocate = 1

    # depending if npot is available, use the real size or pot size
    if make_npot and gl_has_capability(c_GLCAP_NPOT):
        texture_width = width
        texture_height = height
    else:
        texture_width = _nearest_pow2(width)
        texture_height = _nearest_pow2(height)

    # create the texture with the future color format.
    icolorfmt = _convert_gl_format(icolorfmt)
    texture = Texture(texture_width, texture_height, target,
                      colorfmt=colorfmt, bufferfmt=bufferfmt, mipmap=mipmap,
                      callback=callback, icolorfmt=icolorfmt)
    if allocate or make_npot:
        texture.flags |= TI_NEED_ALLOCATE

    # set default parameter for this texture
    texture.set_wrap('clamp_to_edge')
    if mipmap:
        texture.set_min_filter('linear_mipmap_nearest')
        texture.set_mag_filter('linear')
    else:
        texture.set_min_filter('linear')
        texture.set_mag_filter('linear')

    # if the texture size is the same as initial size, return the texture
    # unmodified
    if texture_width == width and texture_height == height:
        return texture

    # otherwise, return a region of that texture
    return texture.get_region(0, 0, width, height)


def texture_create(size=None, colorfmt=None, bufferfmt=None, mipmap=False,
    callback=None, icolorfmt=None):
    '''Create a texture based on size.

    :Parameters:
        `size`: tuple, defaults to (128, 128)
            Size of the texture.
        `colorfmt`: str, defaults to 'rgba'
            Color format of the texture. Can be 'rgba' or 'rgb',
            'luminance' or 'luminance_alpha'. On desktop, additional values are
            available: 'red', 'rg'.
        `icolorfmt`: str, defaults to the value of `colorfmt`
            Internal format storage of the texture. Can be 'rgba' or 'rgb',
            'luminance' or 'luminance_alpha'. On desktop, additional values are
            available: 'r8', 'rg8', 'rgba8'.
        `bufferfmt`: str, defaults to 'ubyte'
            Internal buffer format of the texture. Can be 'ubyte', 'ushort',
            'uint', 'bute', 'short', 'int' or 'float'.
        `mipmap`: bool, defaults to False
            If True, it will automatically generate the mipmap texture.
        `callback`: callable(), defaults to False
            If a function is provided, it will be called when data is
            needed in the texture.

    .. versionchanged:: 1.7.0
        :attr:`callback` has been added
    '''
    cdef int width = 128, height = 128, allocate = 1
    if size is not None:
        width, height = size
    if colorfmt is None:
        colorfmt = 'rgba'
    if bufferfmt is None:
        bufferfmt = 'ubyte'
    if callback is not None:
        allocate = 0
    if icolorfmt is None:
        icolorfmt = colorfmt
    return _texture_create(width, height, colorfmt, bufferfmt, mipmap,
            allocate, callback, icolorfmt)


def texture_create_from_data(im, mipmap=False):
    '''Create a texture from an ImageData class.
    '''
    cdef int width = im.width
    cdef int height = im.height
    cdef int allocate = 1
    cdef int no_blit = 0
    cdef Texture texture

    # optimization, if the texture is power of 2, don't allocate in
    # _texture_create, but allocate in blit_data => only 1 upload
    if _is_pow2(width) and _is_pow2(height):
        allocate = 0
    elif gl_has_capability(c_GLCAP_NPOT):
        allocate = 0

    # if imagedata have more than one image, activate mipmap
    if im.have_mipmap:
        mipmap = True

    if not cgldef.kivy_opengl_es2:
        if gl_get_version_major() < 3:
            mipmap = False

    if width == 0 or height == 0:
        height = width = 1
        allocate = 1
        no_blit = 1
    texture = _texture_create(width, height, im.fmt, 'ubyte', mipmap, allocate,
                             None, im.fmt)
    if texture is None:
        return None

    texture._source = im.source
    if no_blit == 0:
        texture.blit_data(im)

    return texture


cdef class Texture:
    '''Handle an OpenGL texture. This class can be used to create simple
    textures or complex textures based on ImageData.'''

    _sequenced_textures = {}
    '''Internal use only for textures of sequenced images
    '''
    create = staticmethod(texture_create)
    create_from_data = staticmethod(texture_create_from_data)

    def __init__(self, width, height, target, texid=0, colorfmt='rgb',
            bufferfmt='ubyte', mipmap=False, source=None, callback=None,
            icolorfmt='rgb'):
        self.observers = []
        self._width         = width
        self._height        = height
        self._target        = target
        self._id            = texid
        self._mipmap        = mipmap
        self._wrap          = None
        self._min_filter    = None
        self._mag_filter    = None
        self._is_allocated  = 0
        self._uvx           = 0.
        self._uvy           = 0.
        self._uvw           = 1.
        self._uvh           = 1.
        self._colorfmt      = colorfmt
        self._bufferfmt     = bufferfmt
        self._icolorfmt     = icolorfmt
        self._source        = source
        self._nofree        = 0
        self._callback      = callback

        if texid == 0:
            self.flags |= TI_NEED_GEN
        if callback is not None:
            self.flags |= TI_NEED_PIXELS

        self.update_tex_coords()
        get_context().register_texture(self)

    def __dealloc__(self):
        get_context().dealloc_texture(self)

    cdef void update_tex_coords(self):
        self._tex_coords[0] = self._uvx
        self._tex_coords[1] = self._uvy
        self._tex_coords[2] = self._uvx + self._uvw
        self._tex_coords[3] = self._uvy
        self._tex_coords[4] = self._uvx + self._uvw
        self._tex_coords[5] = self._uvy + self._uvh
        self._tex_coords[6] = self._uvx
        self._tex_coords[7] = self._uvy + self._uvh

    def add_reload_observer(self, callback):
        '''Add a callback to be called after the whole graphics context has
        been reloaded. This is where you can reupload your custom data into
        the GPU.

        .. versionadded:: 1.2.0

        :Parameters:
            `callback`: func(context) -> return None
                The first parameter will be the context itself.
        '''
        self.observers.append(WeakMethod(callback))

    def remove_reload_observer(self, callback):
        '''Remove a callback from the observer list, previously added by
        :meth:`add_reload_observer`.

        .. versionadded:: 1.2.0

        '''
        for cb in self.observers[:]:
            method = cb()
            if method is None or method is callback:
                self.observers.remove(cb)
                continue

    cdef void allocate(self):
        cdef int iglfmt, glfmt, iglbufferfmt, datasize, dataerr = 0
        cdef void *data = NULL
        cdef int is_npot = 0

        # check if it's a pot or not
        if not _is_pow2(self._width) or not _is_pow2(self._height):
            make_npot = is_npot = 1

        # prepare information needed for nogil
        glfmt = _color_fmt_to_gl(self._colorfmt)
        iglfmt = _color_fmt_to_gl(self._icolorfmt)
        iglbufferfmt = _buffer_fmt_to_gl(self._bufferfmt)
        datasize = self._width * self._height * \
                _gl_format_size(glfmt) * _buffer_type_to_gl_size(self._bufferfmt)

        # act as we have been able to allocate the texture
        self._is_allocated = 1

        # do the rest outside the Python GIL
        with nogil:
            data = calloc(1, datasize)
            if data != NULL:
                # ensure pixel upload is correct
                _gl_prepare_pixels_upload(self._width)

                # do the initial upload with fake data
                cgl.glTexImage2D(self._target, 0, iglfmt, self._width, self._height,
                        0, glfmt, iglbufferfmt, data)

                # free the data !
                free(data)

                # create mipmap if needed
                if self._mipmap and is_npot == 0:
                    cgl.glGenerateMipmap(self._target)
            else:
                dataerr = 1

        if dataerr:
            self._is_allocated = 0
            raise Exception('Unable to allocate memory for texture (size is %s)' %
                            datasize)

    cpdef flip_vertical(self):
        '''Flip tex_coords for vertical display.'''
        self._uvy += self._uvh
        self._uvh = -self._uvh
        self.update_tex_coords()

    cpdef flip_horizontal(self):
        '''Flip tex_coords for horizontal display.

        .. versionadded:: 1.9.0

        '''
        self._uvx += self._uvw
        self._uvw = -self._uvw
        self.update_tex_coords()

    cpdef get_region(self, x, y, width, height):
        '''Return a part of the texture defined by the rectangular arguments
        (x, y, width, height). Returns a :class:`TextureRegion` instance.'''
        return TextureRegion(x, y, width, height, self)

    def ask_update(self, callback):
        '''Indicate that the content of the texture should be updated and the
        callback function needs to be called when the texture will be
        used.
        '''
        self.flags |= TI_NEED_PIXELS
        self._callback = callback

    cpdef bind(self):
        '''Bind the texture to the current opengl state.'''
        cdef GLuint value

        # if we have no change to apply, just bind and exit
        if not self.flags:
            cgl.glBindTexture(self._target, self._id)
            log_gl_error('Texture.bind-glBindTexture')
            return

        if self.flags & TI_NEED_GEN:
            self.flags &= ~TI_NEED_GEN
            cgl.glGenTextures(1, &self._id)
            log_gl_error('Texture.bind-glGenTextures')

        cgl.glBindTexture(self._target, self._id)
        log_gl_error('Texture.bind-glBindTexture')

        if self.flags & TI_NEED_ALLOCATE:
            self.flags &= ~TI_NEED_ALLOCATE
            self.allocate()

        if self.flags & TI_NEED_PIXELS:
            self.flags &= ~TI_NEED_PIXELS
            if self._callback:
                self._callback(self)
                self._callback = None

        if self.flags & TI_MIN_FILTER:
            self.flags &= ~TI_MIN_FILTER
            value = _str_to_gl_texture_min_filter(self._min_filter)
            cgl.glTexParameteri(self._target, GL_TEXTURE_MIN_FILTER, value)
            log_gl_error('Texture.bind-glTexParameteri (GL_TEXTURE_MIN_FILTER)')

        if self.flags & TI_MAG_FILTER:
            self.flags &= ~TI_MAG_FILTER
            value = _str_to_gl_texture_mag_filter(self._mag_filter)
            cgl.glTexParameteri(self._target, GL_TEXTURE_MAG_FILTER, value)
            log_gl_error('Texture.bind-glTexParameteri (GL_TEXTURE_MAG_FILTER')

        if self.flags & TI_WRAP:
            self.flags &= ~TI_WRAP
            value = _str_to_gl_texture_wrap(self._wrap)
            cgl.glTexParameteri(self._target, GL_TEXTURE_WRAP_S, value)
            log_gl_error('Texture.bind-glTexParameteri (GL_TEXTURE_WRAP_S)')
            cgl.glTexParameteri(self._target, GL_TEXTURE_WRAP_T, value)
            log_gl_error('Texture.bind-glTexParameteri (GL_TEXTURE_WRAP_T')

    cdef void set_min_filter(self, x):
        if self._min_filter != x:
            self._min_filter = x
            self.flags |= TI_MIN_FILTER

    cdef void set_mag_filter(self, x):
        if self._mag_filter != x:
            self._mag_filter = x
            self.flags |= TI_MAG_FILTER

    cdef void set_wrap(self, x):
        if self._wrap != x:
            self._wrap = x
            self.flags |= TI_WRAP

    def blit_data(self, im, pos=None):
        '''Replace a whole texture with image data.
        '''
        blit = self.blit_buffer

        # depending if imagedata have mipmap, think different.
        if not im.have_mipmap:
            blit(im.data, size=im.size, colorfmt=im.fmt, pos=pos,
                    rowlength=im.rowlength)
        else:
            # upload each level
            for level, width, height, data, rowlength in im.iterate_mipmaps():
                blit(data, size=(width, height),
                     colorfmt=im.fmt, pos=pos,
                     mipmap_level=level, mipmap_generation=False,
                     rowlength=rowlength)

    @cython.cdivision(True)
    def blit_buffer(self, pbuffer, size=None, colorfmt=None,
                    pos=None, bufferfmt=None, mipmap_level=0,
                    mipmap_generation=True, int rowlength=0):
        '''Blit a buffer into the texture.

        .. note::

            Unless the canvas will be updated due to other changes,
            :meth:`~kivy.graphics.instructions.Canvas.ask_update` should be
            called in order to update the texture.

        :Parameters:
            `pbuffer`: bytes, or a class that implements the buffer interface\
 (including memoryview).
                A buffer containing the image data. It can be either a bytes
                object or a instance of a class that implements the python
                buffer interface, e.g. `array.array`, `bytearray`, numpy arrays
                etc. If it's not a bytes object, the underlying buffer must
                be contiguous, have only one dimension and must not be
                readonly, even though the data is not modified, due to a cython
                limitation. See module description for usage details.
            `size`: tuple, defaults to texture size
                Size of the image (width, height)
            `colorfmt`: str, defaults to 'rgb'
                Image format, can be one of 'rgb', 'rgba', 'bgr', 'bgra',
                'luminance' or 'luminance_alpha'.
            `pos`: tuple, defaults to (0, 0)
                Position to blit in the texture.
            `bufferfmt`: str, defaults to 'ubyte'
                Type of the data buffer, can be one of 'ubyte', 'ushort',
                'uint', 'byte', 'short', 'int' or 'float'.
            `mipmap_level`: int, defaults to 0
                Indicate which mipmap level we are going to update.
            `mipmap_generation`: bool, defaults to True
                Indicate if we need to regenerate the mipmap from level 0.

        .. versionchanged:: 1.0.7

            added `mipmap_level` and `mipmap_generation`

        .. versionchanged:: 1.9.0
            `pbuffer` can now be any class instance that implements the python
            buffer interface and / or memoryviews thereof.

        '''
        cdef GLuint target = self._target
        cdef int glbufferfmt
        if colorfmt is None:
            colorfmt = 'rgb'
        if bufferfmt is None:
            bufferfmt = 'ubyte'
        if pos is None:
            pos = (0, 0)
        if size is None:
            size = self.size
        glbufferfmt = _buffer_fmt_to_gl(bufferfmt)

        # gles limitation/issue: cannot blit buffer on a different
        # buffer/colorfmt
        # Reference: https://github.com/kivy/kivy/issues/1600
        if gles_limts:
            if colorfmt.lower() != self.colorfmt.lower():
                raise Exception((
                    "GLES LIMIT: Cannot blit with a different colorfmt than "
                    "the created texture. (texture has {}, you passed {}). "
                    "Consider setting KIVY_GLES_LIMITS"
                    ).format(self.colorfmt, colorfmt))
            if bufferfmt.lower() != self.bufferfmt.lower():
                raise Exception((
                    "GLES LIMIT: Cannot blit with a different bufferfmt than "
                    "the created texture. (texture has {}, you passed {}). "
                    "Consider setting KIVY_GLES_LIMITS"
                    ).format(self.bufferfmt, bufferfmt))

        # bind the texture, and create anything that should be created at this
        # time.
        self.bind()

        # need conversion, do check here because it seems to be faster ?
        if not gl_has_texture_native_format(colorfmt):
            pbuffer, colorfmt = convert_to_gl_format(pbuffer, colorfmt,
                                                     size[0], size[1])
        cdef char [:] char_view
        cdef short [:] short_view
        cdef unsigned short [:] ushort_view
        cdef int [:] int_view
        cdef unsigned int [:] uint_view
        cdef float [:] float_view
        cdef char *cdata = NULL
        cdef long datasize = 0
        if isinstance(pbuffer, bytes):  # if it's bytes, just use memory
            cdata = <bytes>pbuffer  # explicit bytes
            datasize = <long>len(pbuffer)
        else:   # if it's a memoryview or buffer type, use start of memory
            if glbufferfmt == GL_UNSIGNED_BYTE or glbufferfmt == GL_BYTE:
                char_view = pbuffer
                cdata = &char_view[0]
                datasize = char_view.nbytes
            elif glbufferfmt == GL_SHORT:
                short_view = pbuffer
                cdata = <char *>&short_view[0]
                datasize = short_view.nbytes
            elif glbufferfmt == GL_UNSIGNED_SHORT:
                ushort_view = pbuffer
                cdata = <char *>&ushort_view[0]
                datasize = ushort_view.nbytes
            elif glbufferfmt == GL_INT:
                int_view = pbuffer
                cdata = <char *>&int_view[0]
                datasize = int_view.nbytes
            elif glbufferfmt == GL_UNSIGNED_INT:
                uint_view = pbuffer
                cdata = <char *>&uint_view[0]
                datasize = uint_view.nbytes
            elif glbufferfmt == GL_FLOAT:
                float_view = pbuffer
                cdata = <char *>&float_view[0]
                datasize = float_view.nbytes

        # prepare nogil
        cdef int iglfmt = _color_fmt_to_gl(self._icolorfmt)
        cdef int glfmt = _color_fmt_to_gl(colorfmt)
        cdef int x = pos[0]
        cdef int y = pos[1]
        cdef int w = size[0]
        cdef int h = size[1]
        cdef int is_allocated = self._is_allocated
        cdef int is_compressed = _is_compressed_fmt(colorfmt)
        cdef int _mipmap_generation = mipmap_generation and self._mipmap
        cdef int _mipmap_level = mipmap_level

        # if there is a pitch/rowlength passed for the texture,
        # determine the alignment needed, and see if GL can handle it on the
        # current platform.
        cdef int bytes_per_pixels = _gl_format_size(glfmt)
        cdef int target_rowlength = w * bytes_per_pixels * _buffer_type_to_gl_size(bufferfmt)
        cdef int need_unpack = rowlength > 0 and rowlength != target_rowlength
        cdef char *cpdata = NULL
        cdef char *cpsrc
        cdef char *cpdst
        cdef int i
        cdef int require_subimage = 0

        # if the hardware doesn't support native unpack, use alternative method.
        if need_unpack and not gl_has_capability(GLCAP_UNPACK_SUBIMAGE):
            require_subimage = 1
            need_unpack = 0

        with nogil:

            if need_unpack:
                # native unpack supported, use it.
                cgl.glPixelStorei(GL_UNPACK_ROW_LENGTH, rowlength / bytes_per_pixels)
                if y != 0:
                    cgl.glPixelStorei(GL_UNPACK_SKIP_ROWS, y)
                if x != 0:
                    cgl.glPixelStorei(GL_UNPACK_SKIP_PIXELS, x)
                _gl_prepare_pixels_upload(rowlength)

            elif require_subimage:
                # make a temporary copy to a format without alignment for upload
                cpsrc = cdata
                cpdst = cpdata = <char *>malloc(target_rowlength * h)
                for i in range(h):
                    memcpy(cpdst, cpsrc, target_rowlength)
                    cpsrc += rowlength
                    cpdst += target_rowlength
                cdata = cpdata
                datasize = target_rowlength * h

            else:
                _gl_prepare_pixels_upload(w)

            if is_compressed:
                cgl.glPixelStorei(GL_UNPACK_ALIGNMENT, 1)
                cgl.glCompressedTexImage2D(target, _mipmap_level, glfmt, w, h, 0,
                        <GLsizei>datasize, cdata)
            elif is_allocated:
                cgl.glTexSubImage2D(target, _mipmap_level, x, y, w, h, glfmt,
                    glbufferfmt, cdata)
            else:
                cgl.glTexImage2D(target, _mipmap_level, iglfmt, w, h, 0, glfmt,
                    glbufferfmt, cdata)

            if _mipmap_generation:
                cgl.glGenerateMipmap(target)

            if need_unpack:
                cgl.glPixelStorei(GL_UNPACK_ROW_LENGTH, 0)
                if y != 0:
                    cgl.glPixelStorei(GL_UNPACK_SKIP_ROWS, 0)
                if x != 0:
                    cgl.glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0)
            elif require_subimage:
                if cpdata != NULL:
                    free(cpdata)

    def _on_proxyimage_loaded(self, image):
        if image is not self._proxyimage:
            return
        self._reload_propagate(image.image.texture)
        self._proxyimage = None

        # the texture might impact something on the drawing, so ask to refresh
        # the window.
        # FIXME: the texture used in BindTexture should ask for a retrigger
        get_context().flag_update_canvas()

    cdef void reload(self):
        cdef Texture texture
        if self._id != <GLuint>-1:
            return
        if self._source is None:
            # manual texture recreation
            texture = texture_create(self.size, self.colorfmt, self.bufferfmt,
                    self.mipmap)
        else:
            source = osource = self._source
            proto = None
            if source.startswith('zip|'):
                proto = 'zip'
                source = source[4:]
            chr = type(source)
            no_cache, filename, mipmap, count = source.split(chr('|'))
            source = chr(u'{}|{}|{}').format(filename, mipmap, count)

            if not proto:
                proto = filename.split(chr(':'), 1)[0]

            if proto in ('http', 'https', 'ftp', 'smb'):
                from kivy.loader import Loader
                self._proxyimage = Loader.image(filename)
                self._id = 0 # FIXME this will point to an invalid texture ...
                self._proxyimage.bind(on_load=self._on_proxyimage_loaded)
                if self._proxyimage.loaded:
                    self._on_proxyimage_loaded(self._proxyimage)
                return

            mipmap = 0 if mipmap == '0' else 1
            if count == '0':
                if proto =='zip' or filename.endswith('.gif'):
                    from kivy.core.image import ImageLoader
                    image = ImageLoader.load(filename, nocache=True, mipmap=mipmap)

                    texture_list = []
                    create_tex = self.create_from_data
                    for data in image._data[1:]:
                        tex = create_tex(data, mipmap=mipmap)
                        texture_list.append(tex)
                    self._sequenced_textures[filename] = texture_list
                else:
                    from kivy.core.image import Image
                    image = Image(filename, nocache=True, mipmap=mipmap)
                texture = image.texture
            else:
                item_no = int(count) - 1
                texture = self._sequenced_textures[filename][item_no]


        self._reload_propagate(texture)


    cdef void _reload_propagate(self, Texture texture):
        # set the same parameters as our current texture
        texture.set_wrap(self.wrap)
        texture.set_min_filter(self.min_filter)
        texture.set_mag_filter(self.mag_filter)
        texture.flags |= TI_MIN_FILTER | TI_MAG_FILTER | TI_WRAP
        texture.uvpos = self.uvpos
        texture.uvsize = self.uvsize

        # ensure the new opengl ID will not get through GC
        texture.bind()
        self._id = texture.id
        texture._nofree = 1

        # then update content again
        for callback in self.observers[:]:
            method = callback()
            if method is None:
                self.observers.remove(callback)
                continue
            method(self)

    def save(self, filename, flipped=True, fmt=None):
        '''Save the texture content to a file. Check
        :meth:`kivy.core.image.Image.save` for more information.

        The flipped parameter flips the saved image vertically, and
        defaults to True.

        .. versionadded:: 1.7.0

        .. versionchanged:: 1.8.0

            Parameter `flipped` added, defaults to True. All the OpenGL Texture
            are read from bottom / left, it need to be flipped before saving.
            If you don't want to flip the image, set flipped to False.

        .. versionchanged:: 1.11.0

            Parameter `fmt` added, to pass the final format to the image provider.
            Used if filename is a BytesIO
        '''
        from kivy.core.image import Image
        return Image(self).save(filename, flipped=flipped, fmt=fmt)

    def __repr__(self):
        return '<Texture hash=%r id=%d size=%r colorfmt=%r bufferfmt=%r source=%r observers=%d>' % (
            id(self), self._id, self.size, self.colorfmt, self.bufferfmt,
            self._source, len(self.observers))

    @property
    def size(self):
        '''Return the (width, height) of the texture (readonly).
        '''
        return (self.width, self.height)

    @property
    def mipmap(self):
        '''Return True if the texture has mipmap enabled (readonly).
        '''
        return self._mipmap

    @property
    def id(self):
        '''Return the OpenGL ID of the texture (readonly).
        '''
        return self._id

    @property
    def target(self):
        '''Return the OpenGL target of the texture (readonly).
        '''
        return self._target

    @property
    def width(self):
        '''Return the width of the texture (readonly).
        '''
        return self._width

    @property
    def height(self):
        '''Return the height of the texture (readonly).
        '''
        return self._height

    @property
    def tex_coords(self):
        '''Return the list of tex_coords (opengl).
        '''
        return (
            self._tex_coords[0],
            self._tex_coords[1],
            self._tex_coords[2],
            self._tex_coords[3],
            self._tex_coords[4],
            self._tex_coords[5],
            self._tex_coords[6],
            self._tex_coords[7])

    @property
    def uvpos(self):
        '''Get/set the UV position inside the texture.
        '''
        return (self._uvx, self._uvy)

    @uvpos.setter
    def uvpos(self, x):
        self._uvx, self._uvy = x
        self.update_tex_coords()

    @property
    def uvsize(self):
        '''Get/set the UV size inside the texture.

        .. warning::
            The size can be negative if the texture is flipped.
        '''
        return (self._uvw, self._uvh)

    @uvsize.setter
    def uvsize(self, x):
        self._uvw, self._uvh = x
        self.update_tex_coords()

    @property
    def colorfmt(self):
        '''Return the color format used in this texture (readonly).

        .. versionadded:: 1.0.7
        '''
        return self._colorfmt

    @property
    def bufferfmt(self):
        '''Return the buffer format used in this texture (readonly).

        .. versionadded:: 1.2.0
        '''
        return self._bufferfmt

    @property
    def min_filter(self):
        '''Get/set the min filter texture. Available values:

        - linear
        - nearest
        - linear_mipmap_linear
        - linear_mipmap_nearest
        - nearest_mipmap_nearest
        - nearest_mipmap_linear

        Check the opengl documentation for more information about the behavior
        of these values :
        http://www.khronos.org/opengles/sdk/docs/man/xhtml/glTexParameter.xml.
        '''
        return self._min_filter

    @min_filter.setter
    def min_filter(self, x):
        self.set_min_filter(x)

    @property
    def mag_filter(self):
        '''Get/set the mag filter texture. Available values:

        - linear
        - nearest

        Check the opengl documentation for more information about the behavior
        of these values :
        http://www.khronos.org/opengles/sdk/docs/man/xhtml/glTexParameter.xml.
        '''
        return self._mag_filter

    @mag_filter.setter
    def mag_filter(self, x):
        self.set_mag_filter(x)

    @property
    def wrap(self):
        '''Get/set the wrap texture. Available values:

        - repeat
        - mirrored_repeat
        - clamp_to_edge

        Check the opengl documentation for more information about the behavior
        of these values :
        http://www.khronos.org/opengles/sdk/docs/man/xhtml/glTexParameter.xml.
        '''
        return self._wrap

    @wrap.setter
    def wrap(self, wrap):
        self.set_wrap(wrap)

    @property
    def pixels(self):
        '''Get the pixels texture, in RGBA format only, unsigned byte. The
        origin of the image is at bottom left.

        .. versionadded:: 1.7.0
        '''
        from kivy.graphics.fbo import Fbo
        return Fbo(size=self.size, texture=self).pixels


cdef class TextureRegion(Texture):
    '''Handle a region of a Texture class. Useful for non power-of-2
    texture handling.'''

    def __init__(self, int x, int y, int width, int height, Texture origin):
        Texture.__init__(self, width, height, origin.target, origin.id)
        self._is_allocated = 1
        self._mipmap = origin._mipmap
        self._colorfmt = origin._colorfmt
        self._bufferfmt = origin._bufferfmt
        self._icolorfmt = origin._icolorfmt
        self.x = x
        self.y = y
        self.owner = origin

        # recalculate texture coordinate
        cdef float origin_u1, origin_v1
        origin_u1 = origin._uvx
        origin_v1 = origin._uvy
        self._uvx = (x / <float>origin._width) * origin._uvw + origin_u1
        self._uvy = (y / <float>origin._height) * origin._uvh + origin_v1
        self._uvw = (width / <float>origin._width) * origin._uvw
        self._uvh = (height / <float>origin._height) * origin._uvh
        self.update_tex_coords()

    def __repr__(self):
        return '<TextureRegion of %r hash=%r id=%d size=%r colorfmt=%r bufferfmt=%r source=%r observers=%d>' % (
            self.owner, id(self), self._id, self.size, self.colorfmt,
            self.bufferfmt, self._source, len(self.observers))

    cdef void reload(self):
        # texture region are reloaded _after_ normal texture
        # so that could work, except if it's a region of region
        # it's safe to retrigger a reload, since the owner texture will be not
        # really reloaded if its id is not -1.
        self.owner.reload()
        self._id = self.owner.id

        # then update content again
        self.bind()
        for callback in self.observers[:]:
            method = callback()
            if method is None:
                self.observers.remove(callback)
                continue
            method(self)

    def ask_update(self, callback):
        # redirect to owner
        self.owner.ask_update(callback)

    cpdef bind(self):
        self.owner.bind()

    @property
    def pixels(self):
        from kivy.graphics.fbo import Fbo
        from kivy.graphics import Color, Rectangle
        fbo = Fbo(size=self.size)
        fbo.clear()
        self.flip_vertical()
        with fbo:
            Color(1, 1, 1)
            Rectangle(size=self.size, texture=self,
                    tex_coords=self.tex_coords)
        fbo.draw()
        self.flip_vertical()
        return fbo.pixels



"""Low level Metrics
====================

Module for low level metrics operations.

"""
__all__ = (
    'dpi2px', 'NUMERIC_FORMATS', 'dispatch_pixel_scale', 'sync_pixel_scale')


cdef float g_dpi = -1
cdef float g_density = -1
cdef float g_fontscale = -1
cdef EventObservers pixel_scale_observers = EventObservers.__new__(
    EventObservers)
"""These observers are dispatched when the dpi/density/font scale changes.
All NumericProperties bind to it, and dispatch their prop in response (if it
causes changes to them).
"""
NUMERIC_FORMATS = ('in', 'px', 'dp', 'sp', 'pt', 'cm', 'mm')
"""The tuple of supported suffixes in Kivy properties that support e.g.
``"10dp"`` type syntax.
"""


def dispatch_pixel_scale(*args):
    """Dispatches all properties that and watch pixel_scale_observers.

    This should be called by Metrics when it changes any of the scaling
    properties.
    """
    from kivy.metrics import Metrics
    global g_dpi, g_density, g_fontscale

    g_dpi = Metrics.dpi
    g_density = Metrics.density
    g_fontscale = Metrics.fontscale
    pixel_scale_observers.dispatch(None, None, None, None, 0)


def sync_pixel_scale(dpi=None, density=None, fontscale=None):
    """Internal method to manually updated the scaling factors when metrics
    changes, before anything has been dispatched.

    Do not call in user code!"""
    global g_dpi, g_density, g_fontscale
    if dpi is not None:
        g_dpi = dpi
    if density is not None:
        g_density = density
    if fontscale is not None:
        g_fontscale = fontscale


cpdef float dpi2px(value, str ext) except *:
    """Converts the value according to the ext."""
    # 1in = 2.54cm = 25.4mm = 72pt = 12pc
    if g_dpi == -1:
        dispatch_pixel_scale()

    # all of these must linearly scale the value (10xx = 10 * 1xx)
    cdef float rv = <float>float(value)
    if ext == 'in':
        return rv * g_dpi
    elif ext == 'px':
        return rv
    elif ext == 'dp':
        return rv * g_density
    elif ext == 'sp':
        return rv * g_density * g_fontscale
    elif ext == 'pt':
        return rv * g_dpi / <float>72.
    elif ext == 'cm':
        return rv * g_dpi / <float>2.54
    elif ext == 'mm':
        return rv * g_dpi / <float>25.4



"""
Tesselator
==========

.. versionadded:: 1.9.0

.. image:: images/tesselator-filled.png
    :align: right
.. image:: images/tesselator-debug.png
    :align: right

.. warning::

    This is experimental and subject to change as long as this warning notice
    is present. Only TYPE_POLYGONS is currently supported.

Tesselator is a library for tesselating polygons, based on
`libtess2 <https://github.com/memononen/libtess2>`_. It renders concave filled
polygons by first tesselating them into convex polygons. It also supports holes.

Usage
-----

First, you need to create a :class:`Tesselator` object and add contours. The
first one is the external contour of your shape and all of the following ones
should be holes::

    from kivy.graphics.tesselator import Tesselator

    tess = Tesselator()
    tess.add_contour([0, 0, 200, 0, 200, 200, 0, 200])
    tess.add_contour([50, 50, 150, 50, 150, 150, 50, 150])

Second, call the :meth:`Tesselator.tesselate` method to compute the points. It
is possible that the tesselator won't work. In that case, it can return
False::

    if not tess.tesselate():
        print("Tesselator didn't work :(")
        return

After the tessellation, you have multiple ways to iterate over the result. The
best approach is using :data:`Tesselator.meshes` to get a format directly usable
for a :class:`~kivy.graphics.Mesh`::

    for vertices, indices in tess.meshes:
        self.canvas.add(Mesh(
            vertices=vertices,
            indices=indices,
            mode="triangle_fan"
        ))

Or, you can get the "raw" result, with just polygons and x/y coordinates with
:meth:`Tesselator.vertices`::

    for vertices in tess.vertices:
        print("got polygon", vertices)

"""

__all__ = ("Tesselator", "WINDING_ODD", "WINDING_NONZERO", "WINDING_POSITIVE",
           "WINDING_NEGATIVE", "TYPE_POLYGONS", "TYPE_BOUNDARY_CONTOURS")

include "common.pxi"
cimport cython
cimport cpython.array
from array import array
from cython cimport view

cdef extern from "tesselator.h":
    ctypedef struct TESStesselator:
        pass
    ctypedef struct TESSalloc:
        pass
    cdef enum TessElementType:
        TESS_POLYGONS
        TESS_CONNECTED_POLYGONS
        TESS_BOUNDARY_CONTOURS
    cdef enum TessWindingRule:
        TESS_WINDING_ODD
        TESS_WINDING_NONZERO
        TESS_WINDING_POSITIVE
        TESS_WINDING_NEGATIVE
        TESS_WINDING_ABS_GEQ_TWO
    TESStesselator *tessNewTess(TESSalloc *)
    void tessDeleteTess(TESStesselator *)
    void tessAddContour(TESStesselator *, int, void *, int, int)
    int tessTesselate(TESStesselator *, int, int, int, int, float *)
    int tessGetVertexCount(TESStesselator *)
    int tessGetElementCount(TESStesselator *)
    float *tessGetVertices(TESStesselator *)
    int *tessGetVertexIndices(TESStesselator *)
    int *tessGetElements(TESStesselator *)


#: Winding enum: ODD
WINDING_ODD = TESS_WINDING_ODD

#: Winding enum: NONZERO
WINDING_NONZERO = TESS_WINDING_NONZERO

#: Winding enum: POSITIVE
WINDING_POSITIVE = TESS_WINDING_POSITIVE

#: Winding enum: NEGATIVE
WINDING_NEGATIVE = TESS_WINDING_NEGATIVE

#: Winding enum: ABS_GET_TWO
WINDING_ABS_GEQ_TWO = TESS_WINDING_ABS_GEQ_TWO

#: Element type enum: POLYGONS
TYPE_POLYGONS = TESS_POLYGONS

#: Element type enum: BOUNDARY_CONTOURS
TYPE_BOUNDARY_CONTOURS = TESS_BOUNDARY_CONTOURS


cdef int TESS_UNDEF = ~(<int>0)


cdef class Tesselator:
    """Tesselator class. See module for more information about the usage.
    """
    def __cinit__(self):
        self.tess = tessNewTess(NULL)

    def __dealloc__(self):
        tessDeleteTess(<TESStesselator *>self.tess)
        self.tess = NULL

    def add_contour(self, points):
        """
        Add a contour to the tesselator. It can be:

        - a list of `[x, y, x2, y2, ...]` coordinates
        - a float array: `array("f", [x, y, x2, y2, ...])`
        - any buffer with floats in it.
        """
        cdef float [:] float_view
        cdef char *cdata
        cdef long datasize
        cdef long count

        if isinstance(points, (tuple, list)):
            float_view = array("f", points)
        else:
            # must be a memory view or a buffer type
            float_view = points

        cdata = <char *>&float_view[0]
        datasize = float_view.nbytes
        self.add_contour_data(cdata, int(len(points) / 2.))

    cpdef int tesselate(
            self, int winding_rule=WINDING_ODD,
            int element_type=TYPE_POLYGONS, int polysize=65535):
        """
        Compute all the contours added with :meth:`add_contour`, and generate
        polygons.

        :Parameters:
            `winding_rule`: enum
                The winding rule classifies a region as inside if its winding
                number belongs to the chosen category. Can be one of
                WINDING_ODD, WINDING_NONZERO, WINDING_POSITIVE,
                WINDING_NEGATIVE, WINDING_ABS_GEQ_TWO. Defaults to WINDING_ODD.
            `element_type`: enum
                The result type, you can generate the polygons with
                TYPE_POLYGONS, or the contours with TYPE_BOUNDARY_CONTOURS.
                Defaults to TYPE_POLYGONS.
        :return: 1 if the tessellation happened, 0 otherwise.
        :rtype: int

        """

        self.element_type = element_type
        self.polysize = polysize
        return tessTesselate(<TESStesselator *>self.tess, winding_rule, element_type,
                            polysize, 2, NULL)

    @property
    def vertex_count(self):
        """Returns the number of vertex generated.

        This is the raw result, however, because the Tesselator format the
        result for you with :data:`meshes` or :data:`vertices` per polygon,
        you'll have more vertices in the result
        """
        return tessGetVertexCount(<TESStesselator *>self.tess)

    @property
    def element_count(self):
        """Returns the number of convex polygon.
        """
        return tessGetElementCount(<TESStesselator *>self.tess)

    @property
    def vertices(self):
        """
        Iterate through the result of the :meth:`tesselate` in order to give
        only a list of `[x, y, x2, y2, ...]` polygons.
        """
        return self.iterate_vertices(0)

    @property
    def meshes(self):
        """
        Iterate through the result of the :meth:`tesselate` to give a result
        that can be easily pushed into Kivy`s Mesh object.

        It's a list of: `[[vertices, indices], [vertices, indices], ...]`.
        The vertices in the format `[x, y, u, v, x2, y2, u2, v2]`.

        Careful, u/v coordinates are the same as x/y.
        You are responsible to change them for texture mapping if you need to.

        You can create Mesh objects like that::

            tess = Tesselator()
            # add contours here
            tess.tesselate()
            for vertices, indices in self.meshes:
                self.canvas.add(Mesh(
                    vertices=vertices,
                    indices=indices,
                    mode="triangle_fan"))
        """
        return self.iterate_vertices(1)

    cdef void add_contour_data(self, void *cdata, int count):
        tessAddContour(<TESStesselator *>self.tess, 2, <void *>cdata, sizeof(float) * 2, count)

    @cython.boundscheck(False)
    cdef iterate_vertices(self, int mode):
        # mode 0: returns for .vertices
        # mode 1: returns for .meshes

        cdef int nelems, i, j, count
        cdef int *poly
        cdef int *elems
        cdef float *verts = <float *>tessGetVertices(<TESStesselator *>self.tess)
        cdef view.array mesh
        cdef float[:] f_mesh
        ret = []
        if self.element_type == TYPE_POLYGONS:
            nelems = tessGetElementCount(<TESStesselator *>self.tess)
            elems = <int *>tessGetElements(<TESStesselator *>self.tess)
            for i in range(nelems):
                poly = &elems[i * self.polysize]

                # first, count the number of vertices in this polygon
                count = 0
                for j in range(self.polysize):
                    if poly[j] == TESS_UNDEF:
                        break
                    count += 1

                # second, wrote it
                if not count:
                    continue

                if mode == 0:
                    # only x/y per polygon
                    mesh = view.array(shape=(count * 2, ),
                                      itemsize=sizeof(float), format="f")
                    f_mesh = mesh
                    for j in range(count):
                        f_mesh[j * 2] = verts[poly[j] * 2]
                        f_mesh[j * 2 + 1] = verts[poly[j] * 2 + 1]
                    ret.append(mesh.memview)

                elif mode == 1:
                    # mode that can fit to actual Kivy Mesh
                    # x, y, u, v
                    mesh = view.array(shape=(count * 4, ),
                                      itemsize=sizeof(float), format="f")
                    f_mesh = mesh
                    for j in range(count):
                        f_mesh[j * 4] = f_mesh[j * 4 + 2] = verts[poly[j] * 2]
                        f_mesh[j * 4 + 1] = f_mesh[j * 4 + 3] = verts[poly[j] * 2 + 1]
                    ret.append((mesh.memview, range(count)))

                else:
                    raise Exception("Invalid mode")
        else:
            # TODO implement TYPE_BOUNDARY_CONTOURS
            raise NotImplementedError()

        return ret
